//==============================================================================
//Copyright(c), Meilhaus Electronic GmbH
//              http://www.meilhaus.de
//              support@meilhaus.de
//
//Source File : uMain.pas
//Description : Example program for using subdevices with Counter 8254
//              - Open main driver (ME-iDS)
//              - Go through the list of all installed ME-iDS devices
//                and find subdevices of type Counter and subtype 8254
//              - 3 counters at once can be selected from the subdevice combobox
//                as the integrated circuit 82C54 always contains 3 counters
//              - The counters can be used independently
//              - On some boards (e.g. ME-1400) counters of the same component
//                can be cascaded without external connections.
//              - Button "Write preload": Writes the preload value to the counter subdevice
//                and stops the timer if enabled
//              - "Single Read": single read from selected subdevice (counter)
//              - "Write & Run": writes the preload value to the subdevice
//                and starts a timer that reads the counter value periodically
//              - "Stop": stops the timer
//              - "Reset": stop the timer and reset the subdevice
//              - Close main driver when program is closed
//==============================================================================
unit uMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls, meIDS;

type
//-------------------------------------------------
{TCounterSubdevice}
  TCounterSubdevice = class(TCollectionItem)
  private
    FDeviceName: string;
    FIndexDevice: Integer;
    FIndexSubdevice: Integer;
    FSubdeviceCaps: Integer;
    FCounterMaxValue: Integer;
  public
    property DeviceName: string read FDeviceName;
    property IndexDevice: Integer read FIndexDevice;
    property IndexSubdevice: Integer read FIndexSubdevice;
    property SubdeviceCaps: Integer read FSubdeviceCaps;
    property CounterMaxValue: Integer read FCounterMaxValue;
  end;

{TCounters8254}
  TCounters8254 = class(TCollection)
  private
    function GetSubdevice(Index: Integer): TCounterSubdevice;
  public
    function AddSubdevice(ADeviceName: string; ADeviceIndex: Integer;
                          ASubdeviceIndex: Integer; ASubdeviceCaps: Integer;
                          ACtrMaxValue: Integer): TCounterSubdevice;
    property Subdevices[Index: Integer]: TCounterSubdevice read GetSubdevice; default;
  end;

//-------------------------------------------------
{TfrmMain}
  TfrmMain = class(TForm)
    btnClearHwInfo: TButton;
    btnClose: TButton;
    btnAbout: TButton;
    memoInfo: TMemo;
    lbl1: TLabel;
    coBoxSubdevs: TComboBox;
    lbl2: TLabel;
    grpBoxCounter1: TGroupBox;
    grpBoxCounter2: TGroupBox;
    grpBoxCounter3: TGroupBox;
    coBoxMode1: TComboBox;
    lbl3: TLabel;
    lbl4: TLabel;
    coBoxClockSource1: TComboBox;
    lbl5: TLabel;
    btnReset1: TButton;
    btnPreload1: TButton;
    btnRead1: TButton;
    editCtrValue1: TEdit;
    lbl6: TLabel;
    btnRun1: TButton;
    btnStop1: TButton;
    lbl7: TLabel;
    coBoxMode2: TComboBox;
    lbl8: TLabel;
    coBoxClockSource2: TComboBox;
    lbl9: TLabel;
    lbl10: TLabel;
    editCtrValue2: TEdit;
    btnPreload2: TButton;
    btnRead2: TButton;
    btnReset2: TButton;
    btnRun2: TButton;
    btnStop2: TButton;
    lbl11: TLabel;
    coBoxMode3: TComboBox;
    lbl12: TLabel;
    coBoxClockSource3: TComboBox;
    lbl13: TLabel;
    lbl14: TLabel;
    editCtrValue3: TEdit;
    btnPreload3: TButton;
    btnRead3: TButton;
    btnReset3: TButton;
    btnStop3: TButton;
    btnRun3: TButton;
    tmr1: TTimer;
    tmr2: TTimer;
    tmr3: TTimer;
    coBoxPreloadValue1: TComboBox;
    coBoxPreloadValue2: TComboBox;
    coBoxPreloadValue3: TComboBox;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btnAboutClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure coBoxSubdevsChange(Sender: TObject);
    procedure btnPreload1Click(Sender: TObject);
    procedure btnRead1Click(Sender: TObject);
    procedure btnReset1Click(Sender: TObject);
    procedure btnPreload2Click(Sender: TObject);
    procedure btnPreload3Click(Sender: TObject);
    procedure btnRead2Click(Sender: TObject);
    procedure btnReset2Click(Sender: TObject);
    procedure btnRead3Click(Sender: TObject);
    procedure btnReset3Click(Sender: TObject);
    procedure btnClearHwInfoClick(Sender: TObject);
    procedure btnRun1Click(Sender: TObject);
    procedure tmr1Timer(Sender: TObject);
    procedure btnStop1Click(Sender: TObject);
    procedure btnStop2Click(Sender: TObject);
    procedure btnStop3Click(Sender: TObject);
    procedure btnRun2Click(Sender: TObject);
    procedure btnRun3Click(Sender: TObject);
    procedure tmr2Timer(Sender: TObject);
    procedure tmr3Timer(Sender: TObject);
    procedure coBoxMode1Change(Sender: TObject);
    procedure coBoxMode2Change(Sender: TObject);
    procedure coBoxMode3Change(Sender: TObject);
    procedure coBoxClockSource1Change(Sender: TObject);
    procedure coBoxClockSource2Change(Sender: TObject);
    procedure coBoxClockSource3Change(Sender: TObject);
  private
    { Private-Deklarationen }
    openErrSuccess: Integer;
    lineNumber: Integer;
    counters8254: TCounters8254;
    procedure CloseMEiDSDriver();
    function MeVersionToString(version: Integer): string;
    procedure DisplayErrorMessage(ErrMsg: string; meError: Integer);
    procedure GetCounter8254Subdevices(numDevices: Integer);
    procedure GetSubdevicesByType(idxDevice: Integer; subdeviceType: Integer; subdeviceSubtype: Integer);
    procedure ChangeCounterSubdevices(ctrGroupIdx: Integer);
    procedure InitControlsCounter0(ctrGroupIdx: Integer);
    procedure InitControlsCounter1(idx: Integer);
    procedure InitControlsCounter2(idx: Integer);
    procedure StopTimer1();
    procedure StopTimer2();
    procedure StopTimer3();
    procedure StopAllTimers();

    procedure FillCounterSources(counterSubdevice: TCounterSubdevice; coBox: TComboBox);
    function SingleWriteCounter(ctrIdx: Integer): Boolean;
    function SingleReadCounter(ctrIdx: Integer; editValue: TEdit): Integer;
    procedure ResetSubdevice(ctrIdx: Integer);
    procedure StopTimer(ctrIdx: Integer);
    procedure AddCoBoxValue(coBox: TComboBox; newStr: string);
    procedure DisplayVersionLibrary();
    procedure DisplayVersionMainDriver();
    procedure DisplaySubdevices();
    function GetDevicePlugged(idxDevice: Integer; var plugged: Integer): Integer;
    function WritePreloadValue(ctrSubdevice: TCounterSubdevice; value: Integer): Integer;
    function ConfigureCounterSubdevice(ctrSubdevice: TCounterSubdevice;
      ctrMode: Integer; ctrClockSource: Integer): Integer;
    function ReadCounterValue(ctrSubdevice: TCounterSubdevice; var value: Integer): Integer;
    procedure AddCountersText(devName: string; idxDevice, idxSubdevice: Integer; numMatched: Integer);
    procedure AddCounterSubdevice(devName: string; idxDevice: Integer; idxSubdevice: Integer);
    procedure DisplayValues(ctrSubdevice: TCounterSubdevice; ctrModeStr: string; ctrClockSrcStr: string; preloadVal: Integer);
  public
    { Public-Deklarationen }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.dfm}

uses meError, uAbout;

const
  ME_IDS_NOTOPENED = -1;

  Ctr8254Mode: array[0..5] of Integer = (ME_SINGLE_CONFIG_CTR_8254_MODE_0,
                                         ME_SINGLE_CONFIG_CTR_8254_MODE_1,
                                         ME_SINGLE_CONFIG_CTR_8254_MODE_2,
                                         ME_SINGLE_CONFIG_CTR_8254_MODE_3,
                                         ME_SINGLE_CONFIG_CTR_8254_MODE_4,
                                         ME_SINGLE_CONFIG_CTR_8254_MODE_5);

//-------------------------------------------------
{TCounters8254}
function TCounters8254.GetSubdevice(Index: Integer): TCounterSubdevice;
begin
  Result := TCounterSubdevice(inherited GetItem(Index));
end;

function TCounters8254.AddSubdevice(ADeviceName: string; ADeviceIndex: Integer;
                                    ASubdeviceIndex: Integer; ASubdeviceCaps: Integer;
                                    ACtrMaxValue: Integer): TCounterSubdevice;
begin
  Result := inherited Add as TCounterSubdevice;
  Result.FDeviceName := ADeviceName;
  Result.FIndexDevice := ADeviceIndex;
  Result.FIndexSubdevice := ASubdeviceIndex;
  Result.FSubdeviceCaps := ASubdeviceCaps;
  Result.FCounterMaxValue := ACtrMaxValue;
end;

//-------------------------------------------------
{TfrmMain}
procedure TfrmMain.FormCreate(Sender: TObject);
begin
  openErrSuccess := ME_IDS_NOTOPENED;
  lineNumber := 0;
  counters8254 := TCounters8254.Create(TCounterSubdevice);
end;

procedure TfrmMain.FormShow(Sender: TObject);
var
  numDevices: Integer;
begin
  memoInfo.Lines.Append('Opening Main driver...');
  openErrSuccess := meOpen(ME_OPEN_NO_FLAGS);

  if openErrSuccess = ME_ERRNO_SUCCESS then
  begin
    DisplayVersionLibrary;
    DisplayVersionMainDriver;
    memoInfo.Lines.Append('');

    meQueryNumberDevices(numDevices);
    memoInfo.Lines.Append('Number of devices found: ' + IntToStr(numDevices));

    GetCounter8254Subdevices(numDevices);
  end else
  begin
    DisplayErrorMessage('Error opening Main driver ', openErrSuccess);
  end;
end;

procedure TfrmMain.GetCounter8254Subdevices(numDevices: Integer);
var
  meError: Integer;
  idxDevice: Integer;
  plugged: Integer;
begin
  if (openErrSuccess = ME_ERRNO_SUCCESS) and (numDevices > 0) then
  begin
    memoInfo.Lines.Add('Retrieving sub-devices of type ME_TYPE_CTR (Counter), subtype ME_SUBTYPE_CTR_8254 (8254)...');
    for idxDevice := 0 to numDevices - 1 do
    begin
      meError := GetDevicePlugged(idxDevice, plugged);
      if (meError = ME_ERRNO_SUCCESS) then
      begin
        if (plugged = ME_PLUGGED_IN) then
        begin
          GetSubdevicesByType(idxDevice, ME_TYPE_CTR, ME_SUBTYPE_CTR_8254);
        end;
      end else
      begin
        DisplayErrorMessage('meQueryInfoDevice Error: ', meError);
      end;
    end;
  end;
  DisplaySubdevices();
end;

procedure TfrmMain.DisplaySubdevices();
begin
  if coBoxSubDevs.Items.Count > 0 then
  begin
    memoInfo.Lines.Append('Number of Counter sub-devices found: ' + IntToStr(coBoxSubDevs.Items.Count * 3));
    memoInfo.Lines.AddStrings(coBoxSubDevs.Items);
    memoInfo.Lines.Add('');
    coBoxSubDevs.ItemIndex := 0;
    ChangeCounterSubdevices(coBoxSubDevs.ItemIndex);
  end else
  begin
    memoInfo.Lines.Append('No Counter sub-devices of subtype 8254 found.');
    CloseMEiDSDriver;
  end;
end;

procedure TfrmMain.GetSubdevicesByType(idxDevice: Integer; subdeviceType: Integer; subdeviceSubtype: Integer);
var
  idxSubdevice: Integer;
  numSubdevices: Integer;
  meError: Integer;
  idxMatchedSubdevice: Integer;
  numMatched: Integer;
  devName: array[0..ME_DEVICE_NAME_MAX_COUNT] of AnsiChar;
begin
  meError := meQueryNameDevice(idxDevice, devName, ME_DEVICE_NAME_MAX_COUNT);
  if meError = ME_ERRNO_SUCCESS then
  begin
    meError := meQueryNumberSubdevices(idxDevice, numSubdevices);
    if meError = ME_ERRNO_SUCCESS then
    begin
      numMatched := 0;
      idxSubdevice := 0;
      while idxSubdevice < numSubdevices do
      begin
        meError := meQuerySubdeviceByType(idxDevice,
                                          idxSubdevice,
                                          subdeviceType,
                                          subdeviceSubtype,
                                          idxMatchedSubdevice);

        if meError = ME_ERRNO_SUCCESS then
        begin
          Inc(numMatched);
          AddCountersText(string(devName), idxDevice, idxMatchedSubdevice, numMatched);
          AddCounterSubdevice(string(devName), idxDevice, idxMatchedSubdevice);
          idxSubdevice := idxMatchedSubdevice + 1;
        end else
        begin
          Break;
        end;
      end;
    end else
    begin
      DisplayErrorMessage('meQueryNameDevice - Error: ', meError);
    end;
  end;
end;

procedure TfrmMain.AddCountersText(devName: string; idxDevice, idxSubdevice: Integer; numMatched: Integer);
var
  countersTxt: string;
begin
  if ((numMatched -1) mod 3) = 2 then
  begin
    countersTxt := string(devName)
                + ', Device ' + IntToStr(idxDevice)
                + ', Subdevices ' + IntToStr(idxSubdevice -2)
                + ', ' + IntToStr(idxSubdevice -1)
                + ', ' + IntToStr(idxSubdevice);
    coBoxSubDevs.Items.Append(countersTxt);
  end;
end;

procedure TfrmMain.AddCounterSubdevice(devName: string; idxDevice: Integer; idxSubdevice: Integer);
var
  caps: Integer;
  ptrArgs: Integer;
begin
  meQuerySubdeviceCaps(idxDevice, idxSubdevice, caps);
  meQuerySubdeviceCapsArgs(idxDevice, idxSubdevice, ME_CAP_CTR_WIDTH, @ptrArgs, 1);

  counters8254.AddSubdevice(string(devName), idxDevice, idxSubdevice, caps, (1 shl ptrArgs) -1);
end;

procedure TfrmMain.DisplayValues(ctrSubdevice: TCounterSubdevice; ctrModeStr: string; ctrClockSrcStr: string; preloadVal: Integer);
begin
  Inc(lineNumber);
  memoInfo.Lines.Add(Format('%.2d', [lineNumber])
                  + ' Write to ' + ctrSubdevice.DeviceName
                  + ',  Device: ' + IntToStr(ctrSubdevice.IndexDevice)
                  + ',  Subdevice: ' + IntToStr(ctrSubdevice.IndexSubdevice)
                  + ',  Counter mode: ' + ctrModeStr
                  + ',  Clock source: ' + ctrClockSrcStr
                  + ',  Preload value: ' + IntToStr(preloadVal));
end;

procedure TfrmMain.DisplayErrorMessage(ErrMsg: string; meError: Integer);
var
  errorMsg: array[0..ME_ERROR_MSG_MAX_COUNT] of AnsiChar;
begin
  meErrorGetMessage(meError, errorMsg, ME_ERROR_MSG_MAX_COUNT);
  memoInfo.lines.add(ErrMsg + IntToStr(meError));
  memoInfo.Lines.Add('ErrorMessage: ' + errorMsg);
end;

procedure TfrmMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  CloseMEiDSDriver();
  counters8254.Free;
end;

procedure TfrmMain.CloseMEiDSDriver();
begin
  StopAllTimers();
  if openErrSuccess = ME_ERRNO_SUCCESS then
  begin
    meClose(ME_CLOSE_NO_FLAGS);
    openErrSuccess := ME_IDS_NOTOPENED;
    coBoxSubDevs.Clear;
    counters8254.Clear;
  end;
end;

procedure TfrmMain.coBoxClockSource1Change(Sender: TObject);
begin
  if (coBoxClockSource1.Tag <> coBoxClockSource1.ItemIndex) then
  begin
    coBoxClockSource1.Tag := coBoxClockSource1.ItemIndex;
    if tmr1.Enabled then
    begin
      StopTimer1();
    end;
  end;
end;

procedure TfrmMain.coBoxClockSource2Change(Sender: TObject);
begin
  if (coBoxClockSource2.Tag <> coBoxClockSource2.ItemIndex) then
  begin
    coBoxClockSource2.Tag := coBoxClockSource2.ItemIndex;
    if tmr2.Enabled then
    begin
      StopTimer2();
    end;
  end;
end;

procedure TfrmMain.coBoxClockSource3Change(Sender: TObject);
begin
  if (coBoxClockSource3.Tag <> coBoxClockSource3.ItemIndex) then
  begin
    coBoxClockSource3.Tag := coBoxClockSource3.ItemIndex;
    if tmr3.Enabled then
    begin
      StopTimer3();
    end;
  end;
end;

procedure TfrmMain.coBoxMode1Change(Sender: TObject);
begin
  if (coBoxMode1.Tag <> coBoxMode1.ItemIndex) then
  begin
    coBoxMode1.Tag := coBoxMode1.ItemIndex;
    if tmr1.Enabled then
    begin
      StopTimer1();
    end;
  end;
end;

procedure TfrmMain.coBoxMode2Change(Sender: TObject);
begin
  if (coBoxMode2.Tag <> coBoxMode2.ItemIndex) then
  begin
    coBoxMode2.Tag := coBoxMode2.ItemIndex;
    if tmr2.Enabled then
    begin
      StopTimer2();
    end;
  end;
end;

procedure TfrmMain.coBoxMode3Change(Sender: TObject);
begin
  if (coBoxMode3.Tag <> coBoxMode3.ItemIndex) then
  begin
    coBoxMode3.Tag := coBoxMode3.ItemIndex;
    if tmr3.Enabled then
    begin
      StopTimer3();
    end;
  end;
end;

procedure TfrmMain.coBoxSubdevsChange(Sender: TObject);
begin
  ChangeCounterSubdevices(coBoxSubdevs.ItemIndex);
end;

procedure TfrmMain.ChangeCounterSubdevices(ctrGroupIdx: Integer);
begin
  if (ctrGroupIdx >= 0) and (coBoxSubdevs.Tag <> ctrGroupIdx) then
  begin
    StopAllTimers();

    InitControlsCounter0(ctrGroupIdx);
    InitControlsCounter1(ctrGroupIdx);
    InitControlsCounter2(ctrGroupIdx);

    coBoxSubdevs.Tag := ctrGroupIdx;
  end;
end;

procedure TfrmMain.InitControlsCounter0(ctrGroupIdx: Integer);
var
  ctrSubdevice: TCounterSubdevice;
begin
    coBoxMode1.ItemIndex := 0;
    coBoxMode1.Tag := 0;
    ctrSubdevice := counters8254[ctrGroupIdx * 3];
    FillCounterSources(ctrSubdevice, coBoxClockSource1);
    coBoxPreloadValue1.Text := IntToStr(ctrSubdevice.CounterMaxValue);
    editCtrValue1.Text := '0';
    grpBoxCounter1.Caption := Format('Subdevice %d', [ctrSubdevice.IndexSubdevice]);
end;

procedure TfrmMain.InitControlsCounter1(idx: Integer);
var
  ctrSubdevice: TCounterSubdevice;
begin
    coBoxMode2.ItemIndex := 0;
    coBoxMode2.Tag := 0;
    ctrSubdevice := counters8254[idx * 3 +1];
    FillCounterSources(ctrSubdevice, coBoxClockSource2);
    coBoxPreloadValue2.Text := IntToStr(ctrSubdevice.CounterMaxValue);
    editCtrValue2.Text := '0';
    grpBoxCounter2.Caption := Format('Subdevice %d', [ctrSubdevice.IndexSubdevice]);
end;

procedure TfrmMain.InitControlsCounter2(idx: Integer);
var
  ctrSubdevice: TCounterSubdevice;
begin
    coBoxMode3.ItemIndex := 0;
    coBoxMode3.Tag := 0;
    ctrSubdevice := counters8254[idx * 3 +2];
    FillCounterSources(ctrSubdevice, coBoxClockSource3);
    coBoxPreloadValue3.Text := IntToStr(ctrSubdevice.CounterMaxValue);
    editCtrValue3.Text := '0';
    grpBoxCounter3.Caption := Format('Subdevice %d', [ctrSubdevice.IndexSubdevice]);
end;

procedure TfrmMain.FillCounterSources(counterSubdevice: TCounterSubdevice; coBox: TComboBox);
begin
  coBox.Clear;
  if (counterSubdevice.SubdeviceCaps and ME_CAPS_CTR_CLK_EXTERNAL) = ME_CAPS_CTR_CLK_EXTERNAL then
  begin
    coBox.AddItem('External', TObject(ME_REF_CTR_EXTERNAL));
  end;
  if (counterSubdevice.SubdeviceCaps and ME_CAPS_CTR_CLK_PREVIOUS) = ME_CAPS_CTR_CLK_PREVIOUS then
  begin
    coBox.AddItem('Previous', TObject(ME_REF_CTR_PREVIOUS));
  end;
  if (counterSubdevice.SubdeviceCaps and ME_CAPS_CTR_CLK_INTERNAL_1MHZ) = ME_CAPS_CTR_CLK_INTERNAL_1MHZ then
  begin
    coBox.AddItem('Internal 1 MHz', TObject(ME_REF_CTR_INTERNAL_1MHZ));
  end;
  if (counterSubdevice.SubdeviceCaps and ME_CAPS_CTR_CLK_INTERNAL_10MHZ) = ME_CAPS_CTR_CLK_INTERNAL_10MHZ then
  begin
    coBox.AddItem('Internal 10 MHz', TObject(ME_REF_CTR_INTERNAL_10MHZ));
  end;
  if coBox.Items.Count > 0 then
  begin
    coBox.ItemIndex := 0;
    coBox.Tag := 0;
  end;
end;

procedure TfrmMain.btnCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TfrmMain.btnPreload1Click(Sender: TObject);
begin
  StopTimer1();
  SingleWriteCounter(0);
end;

procedure TfrmMain.btnPreload2Click(Sender: TObject);
begin
  StopTimer2();
  SingleWriteCounter(1);
end;

procedure TfrmMain.btnPreload3Click(Sender: TObject);
begin
  StopTimer3();
  SingleWriteCounter(2);
end;

procedure TfrmMain.tmr1Timer(Sender: TObject);
var
  meError: Integer;
begin
  meError := SingleReadCounter(0, editCtrValue1);
  if (meError <> ME_ERRNO_SUCCESS) and (meError <> ME_ERRNO_SUBDEVICE_NOT_RUNNING) then
  begin
    StopTimer1();
    DisplayErrorMessage(grpBoxCounter1.Caption + ' meIOSingle Error: ', meError);
  end;
end;

procedure TfrmMain.tmr2Timer(Sender: TObject);
var
  meError: Integer;
begin
  meError := SingleReadCounter(1, editCtrValue2);
  if (meError <> ME_ERRNO_SUCCESS) and (meError <> ME_ERRNO_SUBDEVICE_NOT_RUNNING) then
  begin
    StopTimer2();
    DisplayErrorMessage(grpBoxCounter2.Caption + ' meIOSingle Error: ', meError);
  end;
end;

procedure TfrmMain.tmr3Timer(Sender: TObject);
var
  meError: Integer;
begin
  meError := SingleReadCounter(2, editCtrValue3);
  if (meError <> ME_ERRNO_SUCCESS) and (meError <> ME_ERRNO_SUBDEVICE_NOT_RUNNING) then
  begin
    StopTimer3();
    DisplayErrorMessage(grpBoxCounter3.Caption + ' meIOSingle Error: ', meError);
  end;
end;

procedure TfrmMain.btnRun1Click(Sender: TObject);
begin
  btnRun1.Enabled := False;
  if SingleWriteCounter(0) then
  begin
    tmr1.Enabled := True;
  end else
  begin
    btnRun1.Enabled := True;
  end;
end;

procedure TfrmMain.btnRun2Click(Sender: TObject);
begin
  btnRun2.Enabled := False;
  if SingleWriteCounter(1) then
  begin
    tmr2.Enabled := True;
  end else
  begin
    btnRun2.Enabled := True;
  end;
end;

procedure TfrmMain.btnRun3Click(Sender: TObject);
begin
  btnRun3.Enabled := False;
  if SingleWriteCounter(2) then
  begin
    tmr3.Enabled := True;
  end else
  begin
    btnRun3.Enabled := True;
  end;
end;

procedure TfrmMain.btnStop1Click(Sender: TObject);
begin
  StopTimer1();
end;

procedure TfrmMain.btnStop2Click(Sender: TObject);
begin
  StopTimer2();
end;

procedure TfrmMain.btnStop3Click(Sender: TObject);
begin
  StopTimer3();
end;

procedure TfrmMain.StopTimer(ctrIdx: Integer);
begin
  case ctrIdx of
    0:
      begin
        StopTimer1();
      end;
    1:
      begin
        StopTimer2();
      end;
    2:
      begin
        StopTimer3();
      end;
  else
    StopAllTimers();
  end;
end;

procedure TfrmMain.StopTimer1();
begin
  tmr1.Enabled := False;
  btnRun1.Enabled := True;
end;

procedure TfrmMain.StopTimer2;
begin
  tmr2.Enabled := False;
  btnRun2.Enabled := True;
end;

procedure TfrmMain.StopTimer3;
begin
  tmr3.Enabled := False;
  btnRun3.Enabled := True;
end;

procedure TfrmMain.StopAllTimers();
begin
  StopTimer1();
  StopTimer2();
  StopTimer3();
end;

function TfrmMain.SingleWriteCounter(ctrIdx: Integer): Boolean;
var
  meError: Integer;
  ctrSubdevice: TCounterSubdevice;
  ctrMode: Integer;
  ctrModeStr: string;
  preloadVal: Integer;
  ctrClockSource: Integer;
  ctrClockSrcStr: string;
begin
  Result := False;
  if coBoxSubDevs.ItemIndex >= 0 then
  begin
    try
      ctrSubdevice := counters8254[coBoxSubDevs.ItemIndex * 3 + ctrIdx];
      case ctrIdx of
        0:
          begin
            preloadVal := StrToInt(coBoxPreloadValue1.Text);
            AddCoBoxValue(coBoxPreloadValue1, coBoxPreloadValue1.Text);
            ctrMode := Ctr8254Mode[coBoxMode1.ItemIndex];
            ctrModeStr := coBoxMode1.Text;
            ctrClockSource := Integer(coBoxClockSource1.Items.Objects[coBoxClockSource1.ItemIndex]);
            ctrClockSrcStr := coBoxClockSource1.Text;
          end;
        1:
          begin
            preloadVal := StrToInt(coBoxPreloadValue2.Text);
            AddCoBoxValue(coBoxPreloadValue2, coBoxPreloadValue2.Text);
            ctrMode := Ctr8254Mode[coBoxMode2.ItemIndex];
            ctrModeStr := coBoxMode2.Text;
            ctrClockSource := Integer(coBoxClockSource2.Items.Objects[coBoxClockSource2.ItemIndex]);
            ctrClockSrcStr := coBoxClockSource2.Text;
          end;
        2:
          begin
            preloadVal := StrToInt(coBoxPreloadValue3.Text);
            AddCoBoxValue(coBoxPreloadValue3, coBoxPreloadValue3.Text);
            ctrMode := Ctr8254Mode[coBoxMode3.ItemIndex];
            ctrModeStr := coBoxMode3.Text;
            ctrClockSource := Integer(coBoxClockSource3.Items.Objects[coBoxClockSource3.ItemIndex]);
            ctrClockSrcStr := coBoxClockSource3.Text;
          end;
      else
        Exit;
      end;

      if (preloadVal < 0) or (preloadVal > ctrSubdevice.CounterMaxValue) then
      begin
        raise Exception.Create('Preload value is out of range. Allowed values are from 0...' + IntToStr(ctrSubdevice.CounterMaxValue));
      end;

      meError := ConfigureCounterSubdevice(ctrSubdevice, ctrMode, ctrClockSource);
      if meError <> ME_ERRNO_SUCCESS then
      begin
        DisplayErrorMessage('meIOSingleConfig - Error: ', meError);
        Exit;
      end;

      meError := WritePreloadValue(ctrSubdevice, preloadVal);
      if meError = ME_ERRNO_SUCCESS then
      begin
        DisplayValues(ctrSubdevice, ctrModeStr, ctrClockSrcStr, preloadVal);
        Result := True;
      end else
      begin
        DisplayErrorMessage('meIOSingle - Error: ', meError);
      end;
    except
      on e: Exception do
      begin
        MessageBox(Handle, PChar(e.Message), PChar(Application.Title), MB_OK + MB_ICONSTOP);
      end;
    end;
  end;
end;

function TfrmMain.ConfigureCounterSubdevice(ctrSubdevice: TCounterSubdevice;
  ctrMode: Integer; ctrClockSource: Integer): Integer;
begin
  Result := meIOSingleConfig(ctrSubdevice.IndexDevice,
                              ctrSubdevice.IndexSubdevice,
                              0,
                              ctrMode,
                              ctrClockSource,
                              ME_TRIG_CHAN_DEFAULT,
                              ME_TRIG_TYPE_NONE,
                              ME_VALUE_NOT_USED,
                              ME_IO_SINGLE_CONFIG_NO_FLAGS);
end;

function TfrmMain.WritePreloadValue(ctrSubdevice: TCounterSubdevice; value: Integer): Integer;
var
  io_single: array[0..0] of TMeIOSingle;
begin
  io_single[0].iDevice:= ctrSubdevice.IndexDevice;
  io_single[0].iSubDevice := ctrSubdevice.IndexSubdevice;
  io_single[0].iChannel := 0;
  io_single[0].iDir := ME_DIR_OUTPUT;
  io_single[0].iValue := value;
  io_single[0].iTimeOut := 0;
  io_single[0].iFlags := ME_IO_SINGLE_TYPE_NO_FLAGS;
  io_single[0].iErrno := 0;

  Result := meIOSingle(io_single[0],
                        1,
                        ME_IO_SINGLE_NO_FLAGS);
end;

procedure TfrmMain.btnRead1Click(Sender: TObject);
var
 meError: Integer;
begin
  meError := SingleReadCounter(0, editCtrValue1);
  if (meError <> ME_ERRNO_SUCCESS) then
  begin
    DisplayErrorMessage(grpBoxCounter1.Caption + ' meIOSingle Error: ', meError);
  end;
end;

procedure TfrmMain.btnRead2Click(Sender: TObject);
var
 meError: Integer;
begin
  meError := SingleReadCounter(1, editCtrValue2);
  if (meError <> ME_ERRNO_SUCCESS) then
  begin
    DisplayErrorMessage(grpBoxCounter2.Caption + ' meIOSingle Error: ', meError);
  end;
end;

procedure TfrmMain.btnRead3Click(Sender: TObject);
var
 meError: Integer;
begin
  meError := SingleReadCounter(2, editCtrValue3);
  if (meError <> ME_ERRNO_SUCCESS) then
  begin
    DisplayErrorMessage(grpBoxCounter3.Caption + ' meIOSingle Error: ', meError);
  end;
end;

function TfrmMain.SingleReadCounter(ctrIdx: Integer; editValue: TEdit): Integer;
var
  meError: Integer;
  ctrSubdevice: TCounterSubdevice;
  ctrValue: Integer;
begin
  Result := ME_ERRNO_SUCCESS;
  if (coBoxSubDevs.ItemIndex >= 0) then
  begin
    ctrSubdevice := counters8254[coBoxSubDevs.ItemIndex * 3 + ctrIdx];

    meError := ReadCounterValue(ctrSubdevice, ctrValue);
    if meError = ME_ERRNO_SUCCESS then
    begin
      editValue.Text := IntToStr(ctrValue);
    end else
    begin
      editValue.Text := 'Error ' + IntToStr(meError);
    end;
    Result := meError;
  end;
end;

function TfrmMain.ReadCounterValue(ctrSubdevice: TCounterSubdevice; var value: Integer): Integer;
var
  meError: Integer;
  io_single: array[0..0] of TMeIOSingle;
begin
  io_single[0].iDevice:= ctrSubdevice.IndexDevice;
  io_single[0].iSubDevice := ctrSubdevice.IndexSubdevice;
  io_single[0].iChannel := 0;
  io_single[0].iDir := ME_DIR_INPUT;
  io_single[0].iValue := 0;
  io_single[0].iTimeOut := 0;
  io_single[0].iFlags := ME_IO_SINGLE_TYPE_NO_FLAGS;
  io_single[0].iErrno := 0;

  meError := meIOSingle(io_single[0],
                        1,
                        ME_IO_SINGLE_NO_FLAGS);
  value := io_single[0].iValue;
  Result := meError;
end;

procedure TfrmMain.btnReset1Click(Sender: TObject);
begin
  ResetSubdevice(0);
  editCtrValue1.Text := '0';
end;

procedure TfrmMain.btnReset2Click(Sender: TObject);
begin
  ResetSubdevice(1);
  editCtrValue2.Text := '0';
end;

procedure TfrmMain.btnReset3Click(Sender: TObject);
begin
  ResetSubdevice(2);
  editCtrValue3.Text := '0';
end;

procedure TfrmMain.ResetSubdevice(ctrIdx: Integer);
var
  meError: Integer;
  ctrSubdevice: TCounterSubdevice;
begin
  if (coBoxSubDevs.ItemIndex >= 0) then
  begin
    StopTimer(ctrIdx);
    ctrSubdevice := counters8254[coBoxSubDevs.ItemIndex * 3 + ctrIdx];
    meError := meIOResetSubdevice(ctrSubdevice.IndexDevice,
                                  ctrSubdevice.IndexSubdevice,
                                  ME_IO_RESET_SUBDEVICE_NO_FLAGS
                                  );
    if meError = ME_ERRNO_SUCCESS then
    begin
      memoInfo.Lines.Add('Reset Device ' + IntToStr(ctrSubdevice.IndexDevice)
                     + ', Subdevice ' + IntToStr(ctrSubdevice.IndexSubdevice));
    end else
    begin
      DisplayErrorMessage('meIOResetSubdevice - Error: ', meError);
    end;
  end;
end;

function TfrmMain.GetDevicePlugged(idxDevice: Integer; var plugged: Integer): Integer;
var
  busNo: Integer;
  vendorId: Integer;
  deviceId: Integer;
  funcNo: Integer;
  busType: Integer;
  serialNo: Integer;
  devNo: Integer;
begin
  Result := meQueryInfoDevice(idxDevice,
                    vendorId,
                    deviceId,
                    serialNo,
                    busType,
                    busNo,
                    devNo,
                    funcNo,
                    plugged);
end;

procedure TfrmMain.btnClearHwInfoClick(Sender: TObject);
begin
  memoInfo.Clear;
  lineNumber := 0;
end;

procedure TfrmMain.DisplayVersionLibrary;
var
  ver: Integer;
begin
  meQueryVersionLibrary(ver);
  memoInfo.Lines.Append('Library version: ' + MeVersionToString(ver));
end;

procedure TfrmMain.DisplayVersionMainDriver;
var
  ver: Integer;
begin
  meQueryVersionMainDriver(ver);
  memoInfo.Lines.Append('Main driver version: ' + MeVersionToString(ver));
end;

function TfrmMain.MeVersionToString(version: Integer): string;
begin
  Result := Format('%d.%d.%d.%d', [version shr 24, $FF and (version shr 16), $FF and (version shr 8), $FF and version]);
end;

procedure TfrmMain.AddCoBoxValue(coBox: TComboBox; newStr: string);
var
  tmpStr: string;
begin
  tmpStr := Trim(newStr);
  if (Length(tmpStr) > 0) and (coBox.Items.IndexOf(tmpStr) < 0) then
    coBox.Items.Add(tmpStr);
end;

procedure TfrmMain.btnAboutClick(Sender: TObject);
begin
  frmAbout := TfrmAbout.Create(Self);
  try
    frmAbout.PopupParent := Self;
    frmAbout.ShowModal;
  finally
    frmAbout.Free;
  end;
end;

end.
