//==============================================================================
//Copyright(c), Meilhaus Electronic GmbH
//              http://www.meilhaus.de
//              support@meilhaus.de
//
//Source File : uMain.pas
//Description : Example program for reading property values and property attributes
//              - Open main driver (ME-iDS)
//              - Create the root node (ME-iDS system container) in a TTreeView
//              - Go recursively through the property containers and display all properties found
//              - Add the name and the value of the property to the node text
//              - Add property attributes depending on the data type of the property
//              - Close main driver (ME-iDS) when program is closed
//              - Properties can only be read from boards that are physical present
//              - The application uses the unicode versions of the property functions
//                with wide character strings (WideChar)
//==============================================================================
unit uMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ComCtrls, Vcl.StdCtrls;

type
  TfrmMain = class(TForm)
    tv1: TTreeView;
    btnClose: TButton;
    btnAbout: TButton;
    procedure btnCloseClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btnAboutClick(Sender: TObject);
  private
    { Private-Deklarationen }
    openErrSuccess: Integer;
    procedure DisplayMEiDSProperties();
    procedure DisplayContainerProperties(parentNode: TTreeNode; containerPath: string);
    function GetAccessTypeString(AAccessType: Integer): string;
    function GetPropertyBoolString(ABoolValue: Integer): string;

  public
    { Public-Deklarationen }
  end;

var
  FrmMain: TfrmMain;

implementation

{$R *.dfm}

uses meError, meIDS, meProperties, uAbout;

const
  ME_IDS_NOTOPENED = -1;

procedure TfrmMain.btnCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TfrmMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if (openErrSuccess = ME_ERRNO_SUCCESS) then
  begin
      meClose(ME_CLOSE_NO_FLAGS);
  end;
end;

procedure TfrmMain.FormShow(Sender: TObject);
begin
  openErrSuccess := meOpen(ME_OPEN_NO_FLAGS);
  if openErrSuccess = ME_ERRNO_SUCCESS then
  begin
      DisplayMEiDSProperties();
  end;
end;

procedure TfrmMain.DisplayMEiDSProperties();
var
  treeNode: TTreeNode;
begin
  treeNode := tv1.Items.AddChild(nil, ME_CONTAINER_NAME_ROOT);
  DisplayContainerProperties(treeNode, '\');
  tv1.FullExpand;
  tv1.Selected := treeNode;
  tv1.SetFocus;
end;

procedure TfrmMain.DisplayContainerProperties(parentNode: TTreeNode; containerPath: string);
var
  meError: Integer;
  propertyPath: string;
  propName: string;
  propType: string;
  propStrVal: array[0..256] of WideChar;
  propStrTmpVal: array[0..256] of WideChar;
  propSelectionVal: array[0..256] of WideChar;
  nodeText: string;
  propAccess: string;
  propValue: string;
  numElements: Integer;
  propTypeVal: Integer;
  propIntVal: Integer;
  propBoolVal: Integer;
  propAccType: Integer;
  propIntMin: string;
  propIntMax: string;
  propIntMinVal: Integer;
  propIntMaxVal: Integer;
  propDoubleVal: Double;
  propDoubleMinVal: Double;
  propDoubleMaxVal: Double;
  propDoubleMin: string;
  propDoubleMax: string;
  propIntLength: Integer;
  propStringLength: string;
  propStringMaxLength: string;
  propNumSelections: string;
  propSelection: string;
  propIntNumSelections: Integer;
  i, j: Integer;
  treeNode: TTreeNode;
begin
  propertyPath := containerPath + '\' + ME_ATTRIBUTE_NAME_NUMBER_OF_ELEMENTS;
  meError := mePropertyGetIntW(PWideChar(propertyPath), numElements);
  if (meError = ME_ERRNO_SUCCESS) then
  begin
    for i := 0 to numElements -1 do
    begin
      propName := containerPath + '\' + IntToStr(i);
      meError := mePropertyGetStringW(PWideChar(propName + '\' + ME_ATTRIBUTE_NAME_PROPERTY_NAME), propStrVal, Length(propStrVal));
      if (meError = ME_ERRNO_SUCCESS) then
      begin
        propType := propName + '\' + ME_ATTRIBUTE_NAME_PROPERTY_TYPE;
        meError := mePropertyGetIntW(PWideChar(propType), propTypeVal);
        if (meError = ME_ERRNO_SUCCESS) then
        begin
          nodeText := string(propStrVal);
          if (propTypeVal = ME_PROPERTY_TYPE_CONTAINER) then
          begin
            nodeText := nodeText + ' Type: Container';
            treeNode := tv1.Items.AddChild(parentNode, nodeText);
            DisplayContainerProperties(treeNode, containerPath + '\' + string(propStrVal));
          end else
          begin
            propAccess := containerPath + '\' + propStrVal + '\' + ME_ATTRIBUTE_NAME_PROPERTY_ACCESS;
            meError := mePropertyGetIntW(PChar(propAccess), propAccType);
            if meError = ME_ERRNO_SUCCESS then
            begin
                nodeText := nodeText + GetAccessTypeString(propAccType);
            end;
              propValue := containerPath + '\' + propStrVal;
              if propTypeVal = ME_PROPERTY_TYPE_BOOL then
              begin
                nodeText := nodeText + ', Type: Boolean';
                meError := mePropertyGetIntW(PChar(propValue), propBoolVal);
                if meError = ME_ERRNO_SUCCESS then
                begin
                    nodeText := nodeText + ', Value: ' + GetPropertyBoolString(propBoolVal);
                end;
              end else if propTypeVal = ME_PROPERTY_TYPE_INT then
              begin
                nodeText := nodeText + ', Type: Integer';
                meError := mePropertyGetIntW(PChar(propValue), propIntVal);
                if meError = ME_ERRNO_SUCCESS then
                begin
                   nodeText := nodeText + Format(', Value: %d ($%X)', [propIntVal, propIntVal]);
                end;

                if propAccType = ME_PROPERTY_ACCESS_READ_WRITE then
                begin
                  propIntMin := containerPath + '\' + propStrVal + '\' + ME_ATTRIBUTE_NAME_MIN_VALUE;
                  meError := mePropertyGetIntW(PChar(propIntMin), propIntMinVal);
                  if meError = ME_ERRNO_SUCCESS then
                  begin
                    nodeText := nodeText + ', Minimum: ' + IntToStr(propIntMinVal);
                    propIntMax := containerPath + '\' + propStrVal + '\' + ME_ATTRIBUTE_NAME_MAX_VALUE;
                    meError := mePropertyGetIntW(PChar(propIntMax), propIntMaxVal);
                    if meError = ME_ERRNO_SUCCESS then
                    begin
                      nodeText := nodeText + ', Maximum: ' + IntToStr(propIntMaxVal)
                    end;
                  end;
                end;
              end else if propTypeVal = ME_PROPERTY_TYPE_DOUBLE then
              begin
                nodeText := nodeText + ', Type: Double';
                meError := mePropertyGetDoubleW(PChar(propValue), propDoubleVal);
                if meError = ME_ERRNO_SUCCESS then
                begin
                  nodeText := nodeText + ', Value: ' + FloatToStr(propDoubleVal);
                end;
                if propAccType = ME_PROPERTY_ACCESS_READ_WRITE then
                begin
                  propDoubleMin := containerPath + '\' + propStrVal + '\' + ME_ATTRIBUTE_NAME_MIN_VALUE;
                  meError := mePropertyGetDoubleW(PChar(propDoubleMin), propDoubleMinVal);
                  if meError = ME_ERRNO_SUCCESS then
                  begin
                    nodeText := nodeText + ', Minimum: ' + FloatToStr(propDoubleMinVal);
                    propDoubleMax := containerPath + '\' + propStrVal + '\' + ME_ATTRIBUTE_NAME_MAX_VALUE;
                    meError := mePropertyGetDoubleW(PChar(propDoubleMax), propDoubleMaxVal);
                    if meError = ME_ERRNO_SUCCESS then
                    begin
                      nodeText := nodeText + ', Maximum: ' + FloatToStr(propDoubleMaxVal);
                    end;
                  end;
                end;
              end else if propTypeVal = ME_PROPERTY_TYPE_STRING then
              begin
                nodeText := nodeText + ', Type: String';
                meError := mePropertyGetStringW(PChar(propValue), propStrTmpVal, Length(propStrTmpVal));
                if meError = ME_ERRNO_SUCCESS then
                begin
                  nodeText := nodeText + ', Value: "' + propStrTmpVal + '"';
                end;

                propStringLength := containerPath + '\' + propStrVal + '\' + ME_ATTRIBUTE_NAME_LENGTH;
                meError := mePropertyGetIntW(PChar(propStringLength), propIntLength);
                if meError = ME_ERRNO_SUCCESS then
                begin
                  nodeText := nodeText + ', Length: ' + IntToStr(propIntLength);
                end;

                if propAccType = ME_PROPERTY_ACCESS_READ_WRITE then
                begin
                  propStringMaxLength := containerPath + '\' + propStrVal + '\' + ME_ATTRIBUTE_NAME_MAX_LENGTH;
                  meError := mePropertyGetIntW(PChar(propStringMaxLength), propIntMaxVal);
                  if meError = ME_ERRNO_SUCCESS then
                  begin
                      nodeText := nodeText + ', MaxLength: ' + IntToStr(propIntMaxVal);
                  end;
                end;

              end else if propTypeVal = ME_PROPERTY_TYPE_DEFINE then
              begin
                nodeText := nodeText + ', Type: Define';
                meError := mePropertyGetStringW(PChar(propValue), propStrTmpVal, Length(propStrTmpVal));
                if meError = ME_ERRNO_SUCCESS then
                begin
                  nodeText := nodeText + ', Value: ' + propStrTmpVal;
                end;

                propNumSelections := containerPath + '\' + propStrVal + '\' + ME_ATTRIBUTE_NAME_NUMBER_OF_SELECTIONS;
                meError := mePropertyGetIntW(PChar(propNumSelections), propIntNumSelections);
                if meError = ME_ERRNO_SUCCESS then
                begin
                  nodeText := nodeText + Format(', Selections: %d (', [propIntNumSelections]);
                  for j := 0 to propIntNumSelections -1 do
                  begin
                      propSelection := containerPath + '\' + propStrVal + '\' + ME_ATTRIBUTE_NAME_PREFIX_SELECTION + IntToStr(j);
                      meError := mePropertyGetStringW(PChar(propSelection), propSelectionVal, Length(propSelectionVal));
                      if meError = ME_ERRNO_SUCCESS then
                      begin
                        nodeText := nodeText + propSelectionVal;
                        if j < propIntNumSelections -1 then
                        begin
                          nodeText := nodeText + ', ';
                        end;
                      end;
                  end;
                  nodeText := nodeText + ')';
                end;
              end;
              tv1.Items.AddChild(parentNode, nodeText);
          end;
        end;
      end;

    end;
  end;
end;

function TfrmMain.GetAccessTypeString(AAccessType: Integer): string;
begin
  case AAccessType of
    ME_PROPERTY_ACCESS_READ_ONLY: Result := ', Read only';
    ME_PROPERTY_ACCESS_READ_WRITE:  Result := ', Read / Write';
  else
    Result := 'Error: access type: $' + IntToHex(AAccessType, 8);
  end;
end;

function TfrmMain.GetPropertyBoolString(ABoolValue: Integer): string;
begin
  case ABoolValue of
     0: Result := 'False';
     1: Result := 'True';
  else
    Result := 'Error: boolean value: $' + IntToHex(ABoolValue, 8);
  end;
end;

procedure TfrmMain.btnAboutClick(Sender: TObject);
begin
  frmAbout := TfrmAbout.Create(Self);
  try
    frmAbout.PopupParent := Self;
    frmAbout.ShowModal;
  finally
    frmAbout.Free;
  end;
end;


end.
