//==============================================================================
//Copyright(c), Meilhaus Electronic GmbH
//              http://www.meilhaus.de
//              support@meilhaus.de
//
//Source File : uMain.pas
//Description : Example program for writing single values to Analog Output subdevices
//              - Open main driver (ME-iDS)
//              - Go through the list of all devices
//              - Find AO subdevices and information about channels and ranges
//              - Add AO subdevices to the combobox, so the user can select one
//              - Write data to selected subdevice when user clicks on button "Write Data"
//              - Close main driver when program is closed
//==============================================================================
unit uMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, meIDS, ComCtrls;

type
  TAnalogRange = class;
  TAOSubdevice = class;
  TfrmMain = class(TForm)
    editPhysValue: TEdit;
    lbl1: TLabel;
    btnClearOutput: TButton;
    memoInfo: TMemo;
    btnClose: TButton;
    btnDeviceWrite: TButton;
    lbl3: TLabel;
    coBoxSubdevs: TComboBox;
    grpBox1: TGroupBox;
    btnAbout: TButton;
    coBoxChannel: TComboBox;
    lbl2: TLabel;
    lbl4: TLabel;
    coBoxRange: TComboBox;
    trackbarVal: TTrackBar;
    lbl5: TLabel;
    btnReset: TButton;
    procedure btnCloseClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btnClearOutputClick(Sender: TObject);
    procedure btnDeviceWriteClick(Sender: TObject);
    procedure btnAboutClick(Sender: TObject);
    procedure coBoxSubdevsChange(Sender: TObject);
    procedure coBoxRangeChange(Sender: TObject);
    procedure trackbarValChange(Sender: TObject);
    procedure editPhysValueKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure btnResetClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private-Deklarationen }
    openErrSuccess: Integer;
    subDeviceList: TList;
    lineNumber: Integer;
    function MeVersionToString(version: Integer): string;
    procedure GetAoSubdevices(numDevices: Integer);
    procedure GetSubdevicesByType(idxDevice: Integer; subdeviceType: Integer);
    function GetSubDevTypeName(subdeviceType: Integer): string;
    function GetDevicePlugged(idxDevice: Integer; var plugged: Integer): Integer;
    procedure AddAoSubdevice(devName: string; idxDevice: Integer;
      idxSubdevice: Integer; subdeviceType: Integer);
    procedure AddAnalogRanges(aoSubdevice: TAOSubdevice);
    procedure AddAnalogRange(aoSubdevice: TAOSubdevice; idxRange: Integer);
    procedure FillSubdeviceProperties(idxSelectedSubdev: Integer);
    procedure InitTrackbarPosition(idxSelectedSubdev: Integer);
    procedure CleanupSubdeviceCollection();
    procedure OpenMEiDSDriver();
    procedure CloseMEiDSDriver();
    procedure DisplayVersionLibrary();
    procedure DisplayVersionMainDriver();
    procedure DisplaySubdevices();
    procedure DisplayErrorMessage(ErrMsg: string; meError: Integer);
    function ConfigureSubdeviceForOutput(aoSubdevice: TAOSubdevice;
      idxChannel: Integer; idxRange: Integer): Integer;
    function ConvertPhysicalToDigital(aoSubdevice: TAOSubdevice; idxRange: Integer;
      valPhysical: Double; var valDigital: Integer): Integer;
    function WriteSingleValue(aoSubdevice: TAOSubdevice;
      idxChannel: Integer; valueToWrite: Integer): Integer;
    procedure DisplayValue(aoSubdevice: TAOSubdevice;
      idxRange: Integer; valAnalog: Double; valDigital: Integer);
    procedure FillSubdeviceChannels(aoSubdevice: TAOSubdevice);
    procedure FillSubdeviceRanges(aoSubdevice: TAOSubdevice);
    procedure SetTrackPosition(aoRange: TAnalogRange; physValue: Double);
    function CheckValueMinMax(aoRange: TAnalogRange; valAnalog: Double): Double;
    procedure WritePhysicalValue(valPhysical: Double);

  public
    { Public-Deklarationen }
  end;


//-------------------------------------------------
{TAOSubDevice}
  TAnalogRange = class(TCollectionItem)
  private
    FRangeUnit: Integer;
    FDigitalMax: Integer;
    FPhysicalMin: Double;
    FPhysicalMax: Double;
    FDisplayRange: string;
    FDisplayUnit: string;
  published
    property RangeUnit: Integer read FRangeUnit write FRangeUnit;
    property DigitalMax: Integer read FDigitalMax write FDigitalMax;
    property PhysicalMin: Double read FPhysicalMin write FPhysicalMin;
    property PhysicalMax: Double read FPhysicalMax write FPhysicalMax;
    property DisplayRange: string read FDisplayRange;
    property DisplayUnit: string read FDisplayUnit;
  end;

  TAOSubdevice = class(TCollection)
  private
    FDeviceName: string;
    FIndexDevice: Integer;
    FIndexSubdevice: Integer;
    FNumberOfChannels: Integer;
    function GetRange(Index: Integer): TAnalogRange;
  public
    function AddRange(AUnit: Integer; APhysicalMin, APhysicalMax: Double; ADigitalMax: Integer): TAnalogRange;
    property Ranges[Index: Integer]: TAnalogRange read GetRange;
    property DeviceName: string read FDeviceName;
    property IndexDevice: Integer read FIndexDevice;
    property IndexSubdevice: Integer read FIndexSubdevice;
    property NumberOfChannels: Integer read FNumberOfChannels;
    Constructor Create(ADeviceName: string; AIndexDevice, AIndexSubdevice, ANumberOfChannels: Integer);
  end;
//--End--------------------------------------------

var
  frmMain: TfrmMain;

implementation

uses meError, uAbout;

{$R *.dfm}

const
  ME_IDS_NOTOPENED = -1;
  FORMAT_STR = '0.000';

//----------------------------------------------
{TAOSubDevice}
constructor TAOSubdevice.Create(ADeviceName: string; AIndexDevice, AIndexSubdevice, ANumberOfChannels: Integer);
begin
  inherited Create(TAnalogRange);
  FIndexDevice := AIndexDevice;
  FIndexSubdevice := AIndexSubdevice;
  FNumberOfChannels := ANumberOfChannels;
  FDeviceName := ADeviceName;
end;

function TAOSubdevice.AddRange(AUnit: Integer; APhysicalMin, APhysicalMax: Double; ADigitalMax: Integer): TAnalogRange;
begin
  Result := inherited Add as TAnalogRange;
  Result.FRangeUnit := AUnit;
  Result.FDigitalMax := ADigitalMax;
  Result.FPhysicalMin := APhysicalMin;
  Result.FPhysicalMax := APhysicalMax;

  case AUnit of
    ME_UNIT_VOLT: Result.FDisplayUnit := 'V';
    ME_UNIT_AMPERE: Result.FDisplayUnit := 'A';
    ME_UNIT_HZ: Result.FDisplayUnit := 'Hz';
  else
    Result.FDisplayUnit := 'Unit: $' + IntToHex(AUnit, 8);
  end;
  Result.FDisplayRange := FloatToStr(APhysicalMin) + '...' + FloatToStr(APhysicalMax) + ' ' + Result.FDisplayUnit;
end;

function TAOSubdevice.GetRange(Index: Integer): TAnalogRange;
begin
  Result := TAnalogRange(inherited GetItem(Index));
end;
//----------------------------------------------

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  openErrSuccess := ME_IDS_NOTOPENED;
  lineNumber := 0;
  subDeviceList := TList.Create;
end;

procedure TfrmMain.FormShow(Sender: TObject);
begin
  OpenMEiDSDriver();
end;

procedure TfrmMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  CloseMEiDSDriver();
  subDeviceList.Free;
end;

procedure TfrmMain.OpenMEiDSDriver();
var
  numDevices: Integer;
begin
  memoInfo.Lines.Append('Opening Main driver...');
  openErrSuccess := meOpen(ME_OPEN_NO_FLAGS);

  if openErrSuccess = ME_ERRNO_SUCCESS then
  begin
    DisplayVersionLibrary;
    DisplayVersionMainDriver;
    memoInfo.Lines.Append('');

    meQueryNumberDevices(numDevices);
    memoInfo.Lines.Append('Number of devices found: ' + IntToStr(numDevices));

    GetAoSubdevices(numDevices);
  end else
  begin
    DisplayErrorMessage('Error opening Main driver. Error: ', openErrSuccess);
  end;
end;

procedure TfrmMain.GetAoSubdevices(numDevices: Integer);
var
  meError: Integer;
  idxDevice: Integer;
  plugged: Integer;
begin
  if openErrSuccess = ME_ERRNO_SUCCESS then
  begin
    memoInfo.Lines.Add('Retrieving sub-devices of type ME_TYPE_AO (Analog Out)...');
    for idxDevice := 0 to numDevices - 1 do
    begin
      meError := GetDevicePlugged(idxDevice, plugged);

      if (meError = ME_ERRNO_SUCCESS) and (plugged = ME_PLUGGED_IN) then
      begin
        GetSubdevicesByType(idxDevice, ME_TYPE_AO);
      end else
      begin
        memoInfo.Lines.Append('Device ' + IntToStr(idxDevice) + ' not plugged in.');
      end;
    end;
  end;
  DisplaySubdevices();
end;

procedure TfrmMain.GetSubdevicesByType(idxDevice: Integer; subdeviceType: Integer);
var
  idxSubDevice: Integer;
  numSubDevices: Integer;
  meError: Integer;
  idxMatchedSubdevice: Integer;
  devName: array[0..ME_DEVICE_NAME_MAX_COUNT] of AnsiChar;
begin
  meError := meQueryNumberSubdevices(idxDevice, numSubDevices);
  if meError = ME_ERRNO_SUCCESS then
  begin
    meQueryNameDevice(idxDevice, devName, ME_DEVICE_NAME_MAX_COUNT);
    idxSubDevice := 0;

    while idxSubDevice < numSubDevices do
    begin
      meError := meQuerySubdeviceByType(idxDevice,
                                        idxSubDevice,
                                        subdeviceType,
                                        ME_SUBTYPE_ANY,
                                        idxMatchedSubdevice);

      if meError = ME_ERRNO_SUCCESS then
      begin
        AddAoSubdevice(string(devName), idxDevice, idxMatchedSubdevice, subdeviceType);
        idxSubDevice := idxMatchedSubdevice + 1;
      end else
      begin
        Break;
      end;
    end;
  end;
end;

procedure TfrmMain.AddAoSubdevice(devName: string; idxDevice: Integer; idxSubdevice: Integer; subdeviceType: Integer);
var
  meError: Integer;
  numOfChannels: Integer;
  aoSubdevice: TAOSubdevice;
begin
  meError := meQueryNumberChannels(idxDevice,
                                   idxSubdevice,
                                   numOfChannels);
  if meError = ME_ERRNO_SUCCESS then
  begin
    aoSubdevice := TAOSubdevice.Create(devName, idxDevice, idxSubdevice, numOfChannels);
    subDeviceList.Add(aoSubdevice);
    AddAnalogRanges(aoSubdevice);

    coBoxSubdevs.Items.Append(devName
                              + ',  Device ' + IntToStr(idxDevice)
                              + ',  Subdevice ' + IntToStr(idxSubdevice)
                              + ',  Type: ' + GetSubDevTypeName(subdeviceType)
                              + ',  Channels: ' + IntToStr(numOfChannels)
                              + ',  Ranges: ' + IntToStr(aoSubdevice.Count));
  end;
end;

procedure TfrmMain.AddAnalogRanges(aoSubdevice: TAOSubdevice);
var
  meError: Integer;
  idxRange: Integer;
  numOfRanges: Integer;
begin
  meError := meQueryNumberRanges(aoSubdevice.IndexDevice,
                                 aoSubdevice.IndexSubdevice,
                                 ME_UNIT_ANY,
                                 numOfRanges);
  if meError = ME_ERRNO_SUCCESS then
  begin
    idxRange := 0;
    while idxRange < numOfRanges do
    begin
      AddAnalogRange(aoSubdevice, idxRange);
      Inc(idxRange);
    end;
  end;
end;

procedure TfrmMain.AddAnalogRange(aoSubdevice: TAOSubdevice; idxRange: Integer);
var
  meError: Integer;
  iUnit: Integer;
  physMin: Double;
  physMax: Double;
  digitalMax: Integer;
begin
  meError := meQueryRangeInfo(aoSubdevice.IndexDevice,
                              aoSubdevice.IndexSubdevice,
                              idxRange,
                              iUnit,
                              physMin,
                              physMax,
                              digitalMax);
  if meError = ME_ERRNO_SUCCESS then
  begin
    aoSubdevice.AddRange(iUnit, physMin, physMax, digitalMax);
  end;
end;

procedure TfrmMain.coBoxSubdevsChange(Sender: TObject);
begin
  FillSubdeviceProperties(coBoxSubdevs.ItemIndex);
end;

procedure TfrmMain.FillSubdeviceProperties(idxSelectedSubdev: Integer);
var
  aoSubDevice: TAOSubdevice;
begin
  if coBoxSubdevs.ItemIndex >= 0 then
  begin
    aoSubDevice := TAOSubdevice(subDeviceList[idxSelectedSubdev]);

    FillSubdeviceChannels(aoSubDevice);
    FillSubdeviceRanges(aoSubDevice);
    InitTrackbarPosition(idxSelectedSubdev);
  end;
end;

procedure TfrmMain.FillSubdeviceChannels(aoSubdevice: TAOSubdevice);
var
  i: Integer;
begin
  coBoxChannel.Clear;
  for i := 0 to aoSubdevice.NumberOfChannels - 1 do
  begin
    coBoxChannel.Items.Add('Channel ' + IntToStr(i));
  end;
  if coBoxChannel.Items.Count > 0 then coBoxChannel.ItemIndex := 0;
end;

procedure TfrmMain.FillSubdeviceRanges(aoSubdevice: TAOSubdevice);
var
  i: Integer;
begin
  coBoxRange.Clear;
  for i := 0 to aoSubdevice.Count - 1 do
  begin
    coBoxRange.Items.Add('Range ' + IntToStr(i) + ': ' + aoSubdevice.Ranges[i].DisplayRange);
  end;
  if coBoxRange.Items.Count > 0 then coBoxRange.ItemIndex := 0;
end;

procedure TfrmMain.coBoxRangeChange(Sender: TObject);
begin
  InitTrackbarPosition(coBoxSubdevs.ItemIndex);
end;

procedure TfrmMain.InitTrackbarPosition(idxSelectedSubdev: Integer);
var
  aoSubdevice: TAOSubdevice;
  aoRange: TAnalogRange;
begin
  if idxSelectedSubdev >= 0 then
  begin
    aoSubdevice := TAOSubdevice(subDeviceList[idxSelectedSubdev]);
    aoRange := aoSubdevice.Ranges[coBoxRange.ItemIndex];

    trackbarVal.Min := 0;
    trackbarVal.LineSize := 10;
    trackbarVal.Max := aoRange.DigitalMax;
    trackbarVal.LineSize := aoRange.DigitalMax div 200;
    trackbarVal.PageSize := aoRange.DigitalMax div 40;
    trackbarVal.Frequency := aoRange.DigitalMax div 20;
    SetTrackPosition(aoRange, 0);
    editPhysValue.Text := '0';
  end;
end;

procedure TfrmMain.trackbarValChange(Sender: TObject);
var
  aoSubdevice: TAOSubdevice;
  aoRange: TAnalogRange;
  physValue: Double;
  physRange: Double;
begin
  if coBoxSubdevs.Items.Count > 0 then
  begin
    aoSubdevice := TAOSubdevice(subDeviceList[coBoxSubdevs.ItemIndex]);
    aoRange := aoSubdevice.Ranges[coBoxRange.ItemIndex];
    physRange := aoRange.PhysicalMax - aoRange.PhysicalMin;
    physValue := aoRange.PhysicalMin + (physRange / (trackbarVal.Max - trackbarVal.Min)) * trackbarVal.Position;

    editPhysValue.Text := FormatFloat(FORMAT_STR,physValue);
  end;
end;

procedure TfrmMain.editPhysValueKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  aoSubdevice: TAOSubdevice;
  valAnalog: Double;
  aoRange: TAnalogRange;
begin
  if (Key = VK_RETURN) and (coBoxSubdevs.ItemIndex >= 0)then
  begin
    try
      aoSubdevice := TAOSubdevice(subDeviceList[coBoxSubdevs.ItemIndex]);
      aoRange := aoSubdevice.Ranges[coBoxRange.ItemIndex];
      valAnalog := CheckValueMinMax(aoRange, StrToFloat(editPhysValue.Text));
      SetTrackPosition(aoRange, valAnalog);
    except
      on e: Exception do
      begin
        MessageDlg('Error: ' + e.Message, mtError, [mbOK], 0);
        editPhysValue.SetFocus;
      end;
    end;
  end;
end;

function TfrmMain.CheckValueMinMax(aoRange: TAnalogRange; valAnalog: Double): Double;
begin
  if (valAnalog < aoRange.physicalMin) then
  begin
      valAnalog := aoRange.physicalMin;
      editPhysValue.Text := FormatFloat(FORMAT_STR, valAnalog);
  end else if (valAnalog > aoRange.physicalMax) then
  begin
      valAnalog := aoRange.physicalMax;
      editPhysValue.Text := FormatFloat(FORMAT_STR, valAnalog);
  end;
  Result := valAnalog;
end;

procedure TfrmMain.SetTrackPosition(aoRange: TAnalogRange; physValue: Double);
var
  digitalVal: Integer;
begin
  trackbarVal.OnChange := nil;
  digitalVal := Integer(Round(((physValue - aoRange.PhysicalMin) * aoRange.DigitalMax / (aoRange.PhysicalMax - aoRange.PhysicalMin))));
  trackbarVal.Position := digitalVal;
  Application.ProcessMessages;
  trackbarVal.OnChange := trackbarValChange;
end;

procedure TfrmMain.btnDeviceWriteClick(Sender: TObject);
var
  valPhysical: Double;
begin
  if coBoxSubdevs.ItemIndex >= 0 then
  begin
    try
      valPhysical := StrToFloat(editPhysValue.Text);
      WritePhysicalValue(valPhysical);
    except
      on e: Exception do
      begin
        MessageDlg('Error: ' + e.Message, mtError, [mbOK], 0);
        editPhysValue.SetFocus;
      end;
    end;
  end;
end;

procedure TfrmMain.WritePhysicalValue(valPhysical: Double);
var
  meError: Integer;
  idxChannel: Integer;
  idxRange: Integer;
  valDigital: Integer;
  aoSubdevice: TAOSubdevice;
begin
  idxChannel := coBoxChannel.ItemIndex;
  aoSubdevice := TAOSubdevice(subDeviceList[coBoxSubdevs.ItemIndex]);
  idxRange := coBoxRange.ItemIndex;
  meError := ConvertPhysicalToDigital(aoSubdevice, idxRange, valPhysical, valDigital);
  if meError = ME_ERRNO_SUCCESS then
  begin
    meError := ConfigureSubdeviceForOutput(aoSubdevice, idxChannel, idxRange);
    if meError = ME_ERRNO_SUCCESS then
    begin
      SetTrackPosition(aoSubdevice.Ranges[idxRange], valPhysical);
      meError := WriteSingleValue(aoSubdevice, idxChannel, valDigital);
      if meError = ME_ERRNO_SUCCESS then
      begin
        DisplayValue(aoSubdevice, idxRange, valPhysical, valDigital);
      end;
    end;
    if meError <> ME_ERRNO_SUCCESS then
    begin
      DisplayErrorMessage('meIOSingle - Error: ', meError);
    end;
  end else
  begin
    DisplayErrorMessage('meUtilityPhysicalToDigital - Error: ', meError);
  end;
end;

function TfrmMain.ConvertPhysicalToDigital(aoSubdevice: TAOSubdevice; idxRange: Integer;
  valPhysical: Double; var valDigital: Integer): Integer;
var
  valDigitalMax: Integer;
  dPhysMin: Double;
  dPhysMax: Double;
begin
  dPhysMin := aoSubdevice.Ranges[idxRange].PhysicalMin;
  dPhysMax := aoSubdevice.Ranges[idxRange].PhysicalMax;
  valDigitalMax := aoSubdevice.Ranges[idxRange].DigitalMax;

  if (valPhysical < dPhysMin) then
  begin
    valPhysical := dPhysMin;
  end else if (valPhysical > dPhysMax) then
  begin
    valPhysical := dPhysMax;
  end;

  Result := meUtilityPhysicalToDigital(dPhysMin,
                                       dPhysMax,
                                       valDigitalMax,
                                       valPhysical,
                                       valDigital);
end;

function TfrmMain.ConfigureSubdeviceForOutput(aoSubdevice: TAOSubdevice;
  idxChannel: Integer; idxRange: Integer): Integer;
begin
  Result := meIOSingleConfig(aoSubdevice.IndexDevice,
                aoSubdevice.IndexSubdevice,
                idxChannel,
                idxRange,
                ME_REF_AO_GROUND,
                ME_TRIG_CHAN_DEFAULT,
                ME_TRIG_TYPE_SW,
                ME_VALUE_NOT_USED,
                ME_IO_SINGLE_CONFIG_NO_FLAGS);
end;

function TfrmMain.WriteSingleValue(aoSubdevice: TAOSubdevice;
  idxChannel: Integer; valueToWrite: Integer): Integer;
var
  io_single: array[0..0] of TMeIOSingle;
begin
  io_single[0].iDevice:= aoSubdevice.IndexDevice;
  io_single[0].iSubdevice := aoSubdevice.IndexSubdevice;
  io_single[0].iChannel := idxChannel;
  io_single[0].iDir := ME_DIR_OUTPUT;
  io_single[0].iValue := valueToWrite;
  io_single[0].iTimeOut := 0;
  io_single[0].iFlags := ME_IO_SINGLE_TYPE_NO_FLAGS;
  io_single[0].iErrno := 0;

  Result := meIOSingle(io_single[0],
                         1,
                         ME_IO_SINGLE_NO_FLAGS);
end;

procedure TfrmMain.DisplayValue(aoSubdevice: TAOSubdevice;
  idxRange: Integer; valAnalog: Double; valDigital: Integer);
var
  strUnit: string;
begin
  strUnit := aoSubdevice.Ranges[idxRange].DisplayUnit;
  inc(lineNumber);
  memoInfo.lines.add(Format('%.2d', [lineNumber])
                     + ' Write to ' + aoSubdevice.DeviceName
                     + ',  Device: ' + IntToStr(aoSubdevice.IndexDevice)
                     + ',  Subdevice: ' + IntToStr(aoSubdevice.IndexSubdevice)
                     + ',  Digital value: ' + IntToStr(valDigital)
                     + ',  Physical value: ' + FloatToStr(valAnalog)
                     + ' ' + strUnit);
end;

function TfrmMain.GetDevicePlugged(idxDevice: Integer; var plugged: Integer): Integer;
var
  busNo: Integer;
  vendorId: Integer;
  deviceId: Integer;
  funcNo: Integer;
  busType: Integer;
  serialNo: Integer;
  devNo: Integer;
begin
  Result := meQueryInfoDevice(idxDevice,
                    vendorId,
                    deviceId,
                    serialNo,
                    busType,
                    busNo,
                    devNo,
                    funcNo,
                    plugged);
end;

procedure TfrmMain.DisplaySubdevices();
begin
  if coBoxSubdevs.Items.Count > 0 then
  begin
    memoInfo.Lines.Append('Number of AO sub-devices found: ' + IntToStr(coBoxSubdevs.Items.Count));
    memoInfo.Lines.AddStrings(coBoxSubdevs.Items);
    memoInfo.Lines.Add('');
    coBoxSubdevs.ItemIndex := 0;
    coBoxSubDevsChange(coBoxSubdevs);
    btnDeviceWrite.Enabled := True;
    btnReset.Enabled := btnDeviceWrite.Enabled;
  end else
  begin
    memoInfo.Lines.Append('No AO sub-devices found.');
    CloseMEiDSDriver;
  end;
end;


procedure TfrmMain.CleanupSubdeviceCollection();
var
  i: Integer;
  subDev: TAOSubdevice;
begin
  for i := 0 to subDeviceList.Count -1 do
  begin
    subDev := TAOSubdevice(subDeviceList[i]);
    subDev.Free;
  end;
  subDeviceList.Clear;
end;

procedure TfrmMain.DisplayErrorMessage(ErrMsg: string; meError: Integer);
var
  errorMsg: array[0..ME_ERROR_MSG_MAX_COUNT] of AnsiChar;
begin
  meErrorGetMessage(meError, errorMsg, ME_ERROR_MSG_MAX_COUNT);
  memoInfo.lines.add(ErrMsg + IntToStr(meError));
  memoInfo.Lines.Add('Error Message: ' + errorMsg);
end;

procedure TfrmMain.btnClearOutputClick(Sender: TObject);
begin
  memoInfo.Clear;
  lineNumber := 0;
end;

procedure TfrmMain.btnCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TfrmMain.CloseMEiDSDriver();
begin
  if openErrSuccess = ME_ERRNO_SUCCESS then
  begin
    meClose(ME_CLOSE_NO_FLAGS);
    openErrSuccess := ME_IDS_NOTOPENED;
    CleanupSubdeviceCollection;
    coBoxChannel.Clear;
    coBoxRange.Clear;
    coBoxSubdevs.Clear;
    editPhysValue.Text := '';
    btnDeviceWrite.Enabled := False;
    btnReset.Enabled := False;
  end;
end;

procedure TfrmMain.btnResetClick(Sender: TObject);
var
  meError: Integer;
  aoSubDevice: TAOSubdevice;
begin
  if (coBoxSubdevs.ItemIndex >= 0) then
  begin
    aoSubDevice := TAOSubdevice(subDeviceList[coBoxSubdevs.ItemIndex]);

    meError := meIOResetSubdevice(aoSubDevice.IndexDevice,
                                  aoSubDevice.IndexSubdevice,
                                  ME_IO_RESET_SUBDEVICE_NO_FLAGS);
    if meError = ME_ERRNO_SUCCESS then
    begin
      coBoxChannel.ItemIndex := 0;
      coBoxRange.ItemIndex := 0;
      coBoxRangeChange(coBoxRange);
      memoInfo.Lines.Add('Reset Device ' + IntToStr(aoSubDevice.IndexDevice)
                         + ', Subdevice ' + IntToStr(aoSubDevice.IndexSubdevice));
    end else
    begin
      DisplayErrorMessage('meIOResetSubdevice - Error: ', meError);
    end;
  end;
end;

function TfrmMain.GetSubDevTypeName(subdeviceType: Integer): string;
begin
  case subdeviceType of
    ME_TYPE_INVALID: Result := 'ME_TYPE_INVALID';
    ME_TYPE_AI: Result := 'ME_TYPE_AI';
    ME_TYPE_AO: Result := 'ME_TYPE_AO';
    ME_TYPE_DIO: Result := 'ME_TYPE_DIO';
    ME_TYPE_DO: Result := 'ME_TYPE_DO';
    ME_TYPE_DI: Result := 'ME_TYPE_DI';
    ME_TYPE_CTR: Result := 'ME_TYPE_CTR';
    ME_TYPE_EXT_IRQ: Result := 'ME_TYPE_EXT_IRQ';
    ME_TYPE_FIO: Result := 'ME_TYPE_FIO';
    ME_TYPE_FO: Result := 'ME_TYPE_FO';
    ME_TYPE_FI: Result := 'ME_TYPE_FI';
    ME_TYPE_FPGA: Result := 'ME_TYPE_FPGA';
  else
    Result := '$' + IntToHex(subdeviceType, 8);
  end;
end;

procedure TfrmMain.DisplayVersionLibrary;
var
  ver: Integer;
begin
  meQueryVersionLibrary(ver);
  memoInfo.Lines.Append('Library version: ' + MeVersionToString(ver));
end;

procedure TfrmMain.DisplayVersionMainDriver;
var
  ver: Integer;
begin
  meQueryVersionMainDriver(ver);
  memoInfo.Lines.Append('Main driver version: ' + MeVersionToString(ver));
end;

function TfrmMain.MeVersionToString(version: Integer): string;
begin
  Result := Format('%d.%d.%d.%d', [version shr 24, $FF and (version shr 16), $FF and (version shr 8), $FF and version]);
end;

procedure TfrmMain.btnAboutClick(Sender: TObject);
begin
  frmAbout := TfrmAbout.Create(Self);
  try
    frmAbout.PopupParent := Self;
    frmAbout.ShowModal;
  finally
    frmAbout.Free;
  end;
end;

end.
