﻿//==============================================================================
//Copyright(c), Meilhaus Electronic GmbH
//              http://www.meilhaus.de
//              support@meilhaus.de
//
//Source File : FormMain.cs
//Description : Example program for reading single values from Analog-In (AI) subdevices
//              - Open ME-iDS main driver
//              - Go through the list of all devices and find AI subdevices
//              - Find AI subdevices and information about channels and ranges
//              - Add AI subdevices to the combobox, so the user can select one
//              - Read data from selected subdevice when user clicks on button "Read Data"
//              - Convert digital value to physical value
//              - Close ME-iDS main driver when program is closed
//==============================================================================
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using meIDS_CS;

namespace Single_Read_Analog
{
    public partial class frmMain : Form
    {
        private int openErrSuccess = Constants.ME_IDS_NOTOPENED;
        private int dataLine = 0;
        private List<AiSubdevice> subdeviceList = new List<AiSubdevice>();      

        public frmMain()
        {
            InitializeComponent();
        }

        private void frmMain_Shown(object sender, EventArgs e)
        {
            OpenMEiDSDriver();
        }

        private void frmMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseMEiDSDriver();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void OpenMEiDSDriver()
        {
            try
            {
                textBoxInfo.AppendText("Opening Maindriver..." + Environment.NewLine);
                openErrSuccess = meIDS.meOpen(meIDS.ME_OPEN_NO_FLAGS);      

                if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
                {
                    DisplayVersionLibrary();
                    DisplayVersionMainDriver();
                    textBoxInfo.AppendText(Environment.NewLine);

                    int numDevices;
                    meIDS.meQueryNumberDevices(out numDevices);                 
                    textBoxInfo.AppendText("Number of devices found: " + (numDevices).ToString() + Environment.NewLine);

                    GetAiSubdevices(numDevices);
                }
                else
                {
                    DisplayErrorMessage("meOpen - Error: ", openErrSuccess);
                }
            }
            catch (DllNotFoundException ex)
            {
                textBoxInfo.AppendText("The ME-iDS is not properly installed." + Environment.NewLine);
                textBoxInfo.AppendText(ex.Message + Environment.NewLine);
            }
        }

        private void GetAiSubdevices(int numDevices)
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS && numDevices > 0)
            {
                textBoxInfo.AppendText("Retrieving sub-devices of type ME_TYPE_AI (Analog-Input)..." + Environment.NewLine);
                for (int idxDevice = 0; idxDevice < numDevices; idxDevice++)
                {
                    int plugged;
                    int meError = GetDevicePlugged(idxDevice, out plugged);

                    if (meError == meIDS.ME_ERRNO_SUCCESS && plugged == meIDS.ME_PLUGGED_IN)
                    {
                        GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_AI);  
                    }
                    else
                    {
                        textBoxInfo.AppendText("Device " + idxDevice.ToString() + " not plugged in" + Environment.NewLine);
                    }
                }
            }
            DisplaySubdevices();
        }

        private void GetSubdevicesByType(int idxDevice, int subdeviceType)
        {
            int meError;
            int numSubdevices;
            int idxMatchedSubdevice;
            StringBuilder devName = new StringBuilder(meIDS.ME_DEVICE_NAME_MAX_COUNT);
            meError = meIDS.meQueryNumberSubdevices(idxDevice, out numSubdevices); 
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                int idxSubdevice = 0;
                meIDS.meQueryNameDevice(idxDevice, devName, devName.Capacity);  
                while (idxSubdevice < numSubdevices)
                {
                    meError = meIDS.meQuerySubdeviceByType(idxDevice,                 
                                                           idxSubdevice,              
                                                           subdeviceType,             
                                                           meIDS.ME_SUBTYPE_ANY,      
                                                           out idxMatchedSubdevice);  

                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        AddAiSubdevice(devName.ToString(), idxDevice, idxMatchedSubdevice, subdeviceType);
                        idxSubdevice = idxMatchedSubdevice + 1;                 
                    }
                    else
                    {
                        break;                                                  
                    }
                }
            }
        }

        private int AddAiSubdevice(String devName, int idxDevice, int idxSubdevice, int subdeviceType)
        {
            int numOfChannels;
            int meError = meIDS.meQueryNumberChannels(idxDevice, idxSubdevice, out numOfChannels);
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                AiSubdevice aiSubdev = new AiSubdevice(devName, idxDevice, idxSubdevice, numOfChannels);
                AddAnalogRanges(aiSubdev);
                subdeviceList.Add(aiSubdev);

                coBoxSubDevs.Items.Add(devName.ToString()
                                        + ",  Device " + idxDevice.ToString()
                                        + ",  Subdevice " + idxSubdevice.ToString()
                                        + ",  Type: " + GetSubDevTypeName(subdeviceType)
                                        + ",  Channels: " + numOfChannels.ToString()
                                        + ",  Ranges: " + aiSubdev.GetRangeCount().ToString()
                                        );
            }
            return meError;
        }

        private void AddAnalogRanges(AiSubdevice aiSubdev)
        {
            int numOfRanges;
            int idxRange;
            meIDS.meQueryNumberRanges(aiSubdev.deviceIndex,
                                      aiSubdev.subdevIndex,
                                      meIDS.ME_UNIT_ANY,
                                      out numOfRanges);
            idxRange = 0;
            while (idxRange < numOfRanges)
            {
                AddAnalogRange(aiSubdev, idxRange);
                idxRange++;
            };
        }

        private void AddAnalogRange(AiSubdevice aiSubdev, int idxRange)
        {
            int unit;
            Double physMin;
            Double physMax;
            int digitalMax;

            int meError = meIDS.meQueryRangeInfo(aiSubdev.deviceIndex,
                      aiSubdev.subdevIndex,                  
                      idxRange,                             
                      out unit,                             
                      out physMin,                          
                      out physMax,                          
                      out digitalMax);                      
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                aiSubdev.AddRange(unit, digitalMax, physMin, physMax); 
            };
        }

        private void CloseMEiDSDriver()
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
            {
                meIDS.meClose(meIDS.ME_CLOSE_NO_FLAGS);                                                 
                openErrSuccess = Constants.ME_IDS_NOTOPENED;
                
                btnSingleRead.Enabled = false;
                btnResetSubdev.Enabled = false;
                for (int i = 0; i < subdeviceList.Count; i++)
                {
                    subdeviceList[i].ClearRanges();
                }
                subdeviceList.Clear();
                coBoxChannel.Items.Clear();
                coBoxSubDevs.Items.Clear();
                coBoxRange.Items.Clear();
            }
        }

        private void btnResetSubdev_Click(object sender, EventArgs e)
        {
            AiSubdevice aiSubdev;
            aiSubdev = subdeviceList[coBoxSubDevs.SelectedIndex];
            int meError = meIDS.meIOResetSubdevice(aiSubdev.deviceIndex,
                                                  aiSubdev.subdevIndex,
                                                  meIDS.ME_IO_RESET_SUBDEVICE_NO_FLAGS);
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                textBoxInfo.AppendText("Reset Device " + aiSubdev.deviceIndex.ToString() + ", Subdevice " + aiSubdev.subdevIndex.ToString() + Environment.NewLine);
            }
            else
            {
                DisplayErrorMessage("Error resetting Subdevice", meError);
            }
        }

        private void btnSingleRead_Click(object sender, EventArgs e)
        {
            if (coBoxSubDevs.SelectedIndex >= 0 && coBoxRange.SelectedIndex >= 0)
            {
                int meError;
                AiSubdevice aiSubdevice = subdeviceList[coBoxSubDevs.SelectedIndex]; 
                int idxChannel = coBoxChannel.SelectedIndex;
                int idxRange = coBoxRange.SelectedIndex;

                meError = ConfigureSubdeviceForInput(aiSubdevice, idxChannel, idxRange);
                if (meError == meIDS.ME_ERRNO_SUCCESS)
                {
                    int valDigital;
                    meError = ReadSingleValue(aiSubdevice, idxChannel, out valDigital);

                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        Double valAnalog;
                        meError = ConvertDigitalToPhysical(aiSubdevice, idxRange, valDigital, out valAnalog);

                        if (meError == meIDS.ME_ERRNO_SUCCESS)
                        {
                            DisplayValue(aiSubdevice, idxChannel, coBoxRange.SelectedIndex, valDigital, valAnalog);
                        } 
                        else
                        {
                            DisplayErrorMessage("meUtilityDigitalToPhysical - Error: ", meError);
                        }
                    } 
                    else
                    {
                        DisplayErrorMessage("meIOSingle - Error: ", meError);
                    }
                }
                else
                {
                    DisplayErrorMessage("meIOSingleConfig - Error: ", meError);
                }
            }
        }

        private int ConvertDigitalToPhysical(AiSubdevice aiSubdevice, int idxRange, int value, out Double valAnalog)
        {            
            Double physMin = aiSubdevice.GetAnalogRange(idxRange).physicalMin;
            Double physMax = aiSubdevice.GetAnalogRange(idxRange).physicalMax;
            int digitalMax = aiSubdevice.GetAnalogRange(idxRange).digitalMax;
            int meError = meIDS.meUtilityDigitalToPhysical(physMin,		          
                                                    physMax,                  
                                                    digitalMax,               
                                                    value,
                                                    meIDS.ME_VALUE_NOT_USED,  
                                                    0.0,                      
                                                    out valAnalog);
            return meError;
        }

        private static int ConfigureSubdeviceForInput(AiSubdevice aiSubdevice, int idxChannel, int idxRange)
        {
            int meError;
            meError = meIDS.meIOSingleConfig(aiSubdevice.deviceIndex,
                                      aiSubdevice.subdevIndex, 
                                      idxChannel,              
                                      idxRange,                
                                      meIDS.ME_REF_AI_GROUND,  
                                      meIDS.ME_TRIG_CHAN_DEFAULT,
                                      meIDS.ME_TRIG_TYPE_SW,     
                                      meIDS.ME_VALUE_NOT_USED,   
                                      meIDS.ME_IO_SINGLE_CONFIG_NO_FLAGS);
            return meError;
        }

        private int ReadSingleValue(AiSubdevice aiSubdevice, int idxChannel, out int value)
        {
            meIDS.meIOSingle_t[] io_single = new meIDS.meIOSingle_t[1]; 
            io_single[0].iDevice = aiSubdevice.deviceIndex;
            io_single[0].iSubdevice = aiSubdevice.subdevIndex;
            io_single[0].iChannel = idxChannel;                         
            io_single[0].iDir = meIDS.ME_DIR_INPUT;                     
            io_single[0].iValue = 0;                                    
            io_single[0].iTimeOut = 0;                                  
            io_single[0].iFlags = meIDS.ME_IO_SINGLE_TYPE_NO_FLAGS;     
            io_single[0].iErrno = 0;                                    

            int meError = meIDS.meIOSingle(ref io_single[0],                
                                       1,                               
                                       meIDS.ME_IO_SINGLE_NO_FLAGS);    
            value = io_single[0].iValue;
            return meError;
        }

        private void DisplayValue(AiSubdevice aiSubdevice, int idxChannel, int idxRange, int valDigital, Double valPhysical)
        {
            dataLine++;
            textBoxInfo.AppendText(String.Format("{0:00}", dataLine)
                                + " Read from " + aiSubdevice.deviceName
                                + ",  Device: " + aiSubdevice.deviceIndex.ToString()
                                + ",  Subdevice: " + aiSubdevice.subdevIndex.ToString()
                                + ",  Channel: " + idxChannel.ToString()
                                + ",  Digital value: " + valDigital.ToString()
                                + ",  Physical value: " + String.Format("{0:0.000}", valPhysical)
                                + " " + aiSubdevice.GetAnalogRange(idxRange).displayUnit
                                + Environment.NewLine);
        }

        private void coBoxSubDevs_SelectedIndexChanged(object sender, EventArgs e)
        {
            FillSubdeviceProperties(coBoxSubDevs.SelectedIndex);
        }

        private void FillSubdeviceProperties(int idxSelectedSubdev)
        {
            if (idxSelectedSubdev >= 0)
            {
                AiSubdevice aiSubdevice = subdeviceList[idxSelectedSubdev];
                FillSubdeviceChannels(aiSubdevice);
                FillSubdeviceRanges(aiSubdevice);
            }
        }

        private void FillSubdeviceChannels(AiSubdevice aiSubdevice)
        {
            coBoxChannel.Items.Clear();
            for (int i = 0; i < aiSubdevice.numOfChannels; i++)
            {
                coBoxChannel.Items.Add("Channel " + i.ToString());
            };
            if (coBoxChannel.Items.Count > 0) coBoxChannel.SelectedIndex = 0;
        }

        private void FillSubdeviceRanges(AiSubdevice aiSubdevice)
        {
            AnalogRange aiRange;
            coBoxRange.Items.Clear();
            for (int i = 0; i < aiSubdevice.GetRangeCount(); i++)
            {
                aiRange = aiSubdevice.GetAnalogRange(i);
                coBoxRange.Items.Add("Range " + i.ToString()
                                    + ": " + aiRange.displayRange
                );
            }
            if (coBoxRange.Items.Count > 0) coBoxRange.SelectedIndex = 0;
        }

        private void DisplaySubdevices()
        {
            if (coBoxSubDevs.Items.Count > 0)
            {
                textBoxInfo.AppendText("Number of Analog-Input sub-devices found: " + coBoxSubDevs.Items.Count.ToString() + Environment.NewLine);
                for (int i = 0; i < coBoxSubDevs.Items.Count; i++)
                {
                    textBoxInfo.AppendText(coBoxSubDevs.Items[i] + Environment.NewLine);
                }
                textBoxInfo.AppendText(Environment.NewLine);
                coBoxSubDevs.SelectedIndex = 0;
                btnSingleRead.Enabled = true;
                btnResetSubdev.Enabled = true;
            }
            else
            {
                textBoxInfo.AppendText("No Analog-Input sub-devices found." + Environment.NewLine);
                CloseMEiDSDriver();
            }
        }

        private void DisplayErrorMessage(string message, int meError)
        {
            StringBuilder errMessage = new StringBuilder(meIDS.ME_ERROR_MSG_MAX_COUNT);
            textBoxInfo.AppendText(message + meError.ToString() + Environment.NewLine);
            meIDS.meErrorGetMessage(meError, errMessage, errMessage.Capacity);             
            textBoxInfo.AppendText("ErrorMessage: " + errMessage + Environment.NewLine);    
        }

        private int GetDevicePlugged(int idxDevice, out int plugged)
        {
            int vendorId;
            int deviceId;
            int serialNo;
            int busType;
            int busNo;
            int devNo;
            int funcNo;

            int meError = meIDS.meQueryInfoDevice(idxDevice,
                                             out vendorId,
                                             out deviceId,
                                             out serialNo,
                                             out busType,
                                             out busNo,
                                             out devNo,
                                             out funcNo,
                                             out plugged);              //Indicates whether a device is physically available
            return meError;
        }

        private string GetSubDevTypeName(int subdeviceType)
        {
            string tmpType;
            switch (subdeviceType)
            {
                case meIDS.ME_TYPE_INVALID:
                    tmpType = "ME_TYPE_INVALID";
                    break;
                case meIDS.ME_TYPE_AI:
                    tmpType = "ME_TYPE_AI";
                    break;
                case meIDS.ME_TYPE_AO:
                    tmpType = "ME_TYPE_AO";
                    break;
                case meIDS.ME_TYPE_DIO:
                    tmpType = "ME_TYPE_DIO";
                    break;
                case meIDS.ME_TYPE_DO:
                    tmpType = "ME_TYPE_DO";
                    break;
                case meIDS.ME_TYPE_DI:
                    tmpType = "ME_TYPE_DI";
                    break;
                case meIDS.ME_TYPE_CTR:
                    tmpType = "ME_TYPE_CTR";
                    break;
                case meIDS.ME_TYPE_EXT_IRQ:
                    tmpType = "ME_TYPE_EXT_IRQ";
                    break;
                case meIDS.ME_TYPE_FIO:
                    tmpType = "ME_TYPE_FIO";
                    break;
                case meIDS.ME_TYPE_FO:
                    tmpType = "ME_TYPE_FO";
                    break;
                case meIDS.ME_TYPE_FI:
                    tmpType = "ME_TYPE_FI";
                    break;
                case meIDS.ME_TYPE_FPGA:
                    tmpType = "ME_TYPE_FPGA";
                    break;
                default:
                    tmpType = String.Format("0x{0:X}", subdeviceType);
                    break;
            };
            return tmpType;
        }

        private void DisplayVersionMainDriver()
        {
            int tmpVersion;
            meIDS.meQueryVersionMainDriver(out tmpVersion);
            textBoxInfo.AppendText("Maindriver version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
        }

        private void DisplayVersionLibrary()
        {
            int tmpVersion;
            meIDS.meQueryVersionLibrary(out tmpVersion);
            textBoxInfo.AppendText("Library version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
        }

        private string MeVersionToString(int version)
        {
            return String.Format("{0}.{1}.{2}.{3}", version >> 24, 0xFF & (version >> 16), 0xFF & (version >> 8), 0xFF & version);
        }

        private void btnAbout_Click(object sender, EventArgs e)
        {
            AboutBox1 aboutBox = new AboutBox1();
            aboutBox.ShowDialog();
        }

        private void btnClearHwInfo_Click(object sender, EventArgs e)
        {
            textBoxInfo.Clear();
            dataLine = 0;
        }
    }

    class AnalogRange
    {
        //store characteristics of an analog range
        public int rangeUnit;
        public int digitalMax;
        public Double physicalMin;
        public Double physicalMax;
        public string displayRange;
        public string displayUnit;
        public AnalogRange(int RangeUnit, int DigitalMax, Double PhysicalMin, Double PhysicalMax)
        {
            this.rangeUnit = RangeUnit;
            this.digitalMax = DigitalMax;
            this.physicalMin = PhysicalMin;
            this.physicalMax = PhysicalMax;
            this.SetRangeUnitTexts();
        }

        private void SetRangeUnitTexts()
        {
            Double tmpValMin = Math.Abs(physicalMin);
            Double tmpValMax = Math.Abs(physicalMin);
            StringBuilder sbRange = new StringBuilder();

            if (tmpValMin == 0)
            {
                sbRange.Append("0");
            }
            else if (tmpValMin < 0.001)
            {
                sbRange.Append(string.Format("{0:0.00}", physicalMin * 1000000));
                sbRange.Append("µ");
            }
            else if (tmpValMin < 1)
            {
                sbRange.Append(string.Format("{0:0.00}", physicalMin * 1000));
                sbRange.Append("m");
            }
            else
            {
                sbRange.Append(string.Format("{0:0.00}", Math.Round(physicalMin, 2)));
            }

            if (tmpValMax == 0)
            {
                sbRange.Append("...0");
            }
            else if (tmpValMax < 0.001)
            {
                sbRange.Append(string.Format("...{0:0.00}", physicalMax * 1000000));
                sbRange.Append("µ");
            }
            else if (tmpValMax < 1)
            {
                sbRange.Append(string.Format("...{0:0.00}", physicalMax * 1000));
                sbRange.Append("m");
            }
            else
            {
                sbRange.Append(string.Format("...{0:0.00}", Math.Round(physicalMax, 2)));
            }

            if (rangeUnit == meIDS.ME_UNIT_VOLT)
            {
                displayUnit = "Volt";
                sbRange.Append("V");
            }
            else if (rangeUnit == meIDS.ME_UNIT_AMPERE)
            {
                displayUnit = "Ampere";
                sbRange.Append("A");
            }

            displayRange = sbRange.ToString();  //gives e.g. -10...10V
        }
    }

    class AiSubdevice
    {
        public string deviceName;
        public int deviceIndex;
        public int subdevIndex;
        public int numOfChannels;
        private List<AnalogRange> analogRanges = new List<AnalogRange>();                
        public AiSubdevice(string DeviceName, int DeviceIndex, int SubdevIndex, int NumOfChannels)
        {
            this.deviceName = DeviceName;
            this.deviceIndex = DeviceIndex;
            this.subdevIndex = SubdevIndex;
            this.numOfChannels = NumOfChannels;
        }

        public void AddRange(int RangeUnit, int DigitalMax, Double PhysicalMin, Double PhysicalMax)
        {
            AnalogRange aoRng = new AnalogRange(RangeUnit, DigitalMax, PhysicalMin, PhysicalMax);
            analogRanges.Add(aoRng);
        }

        public int GetRangeCount()
        {
            return analogRanges.Count;
        }

        public AnalogRange GetAnalogRange(int index)
        {
            return analogRanges[index];
        }

        public void ClearRanges()
        {
            analogRanges.Clear();
        }
    }

    static class Constants
    {
        public const int ME_IDS_NOTOPENED = -1;
    }
}
