﻿//==============================================================================
//Copyright(c), Meilhaus Electronic GmbH
//              http://www.meilhaus.de
//              support@meilhaus.de
//
//Source File : FormMain.cs
//Description : Example program for reading single values from DI and DIO sub-devices
//              - Open ME-iDS main driver when program is started
//              - Go through the list of all devices and find DI and DIO subdevices
//              - Add DI and DIO devices to the combobox
//              - Read data from selected subdevice when user clicks on button "Read Data"
//              - Data are always read from channel 0 in this example
//              - Close ME-iDS main driver when program is closed
//==============================================================================
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using meIDS_CS;

namespace Single_Read_Digital
{
    
    public partial class frmMain : Form
    {
        private int openErrSuccess = Constants.ME_IDS_NOTOPENED;
        private int dataLine = 0;
        private List<DiSubdevice> subdeviceList = new List<DiSubdevice>();      

        public frmMain()
        {
            InitializeComponent();
        }

        private void frmMain_Shown(object sender, EventArgs e)
        {
            OpenMEiDSDriver();
        }

        private void OpenMEiDSDriver()
        {
            try
            {
                textBoxInfo.AppendText("Opening Maindriver..." + Environment.NewLine);
                openErrSuccess = meIDS.meOpen(meIDS.ME_OPEN_NO_FLAGS);          

                if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
                {
                    DisplayVersionLibrary();
                    DisplayVersionMainDriver();
                    textBoxInfo.AppendText(Environment.NewLine);

                    int numDevices;
                    meIDS.meQueryNumberDevices(out numDevices);                     
                    textBoxInfo.AppendText("Number of devices found: " + numDevices.ToString() + Environment.NewLine);

                    GetDiDioSubdevices(numDevices);
                }
                else
                {
                    DisplayErrorMessage("meOpen - Error: ", openErrSuccess);
                }
            }
            catch (DllNotFoundException ex)
            {
                textBoxInfo.AppendText("The ME-iDS is not properly installed." + Environment.NewLine);
                textBoxInfo.AppendText(ex.Message + Environment.NewLine);
            }
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void frmMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseMEiDSDriver();
        }

        private void GetDiDioSubdevices(int numDevices)
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS && numDevices > 0)
            {
                textBoxInfo.AppendText("Retrieving subdevices of type ME_TYPE_DI (Digital In) and ME_TYPE_DIO (Digital In/Out)..." + Environment.NewLine);
                for (int idxDevice = 0; idxDevice < numDevices; idxDevice++)
                {
                    int plugged;
                    int meError = GetDevicePlugged(idxDevice, out plugged);
                    if (meError == meIDS.ME_ERRNO_SUCCESS && plugged == meIDS.ME_PLUGGED_IN)
                    {
                        GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_DI);  
                        GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_DIO); 
                    }
                    else
                    {
                        textBoxInfo.AppendText("Device " + idxDevice.ToString() + " not plugged in or returned error." + Environment.NewLine);
                    }
                }
            }
            DisplaySubdevices();
        }

        private void DisplaySubdevices()
        {
            if (coBoxSubDevs.Items.Count > 0)
            {
                textBoxInfo.AppendText("Number of DI or DIO subdevices found: " + coBoxSubDevs.Items.Count.ToString() + Environment.NewLine);
                for (int i = 0; i < coBoxSubDevs.Items.Count; i++)
                {
                    textBoxInfo.AppendText(coBoxSubDevs.Items[i] + Environment.NewLine);  
                }
                textBoxInfo.AppendText(Environment.NewLine);
                coBoxSubDevs.SelectedIndex = 0;                             
                btnSingleRead.Enabled = true;
                btnResetSubdev.Enabled = true;
            }
            else
            {
                textBoxInfo.AppendText("No DI or DIO subdevices found." + Environment.NewLine);  
                CloseMEiDSDriver();
            }
        }

        private void GetSubdevicesByType(int idxDevice, int subdeviceType)
        {
            int meError;
            int numSubdevices;
            int idxMatchedSubdevice;
            StringBuilder devName = new StringBuilder(meIDS.ME_DEVICE_NAME_MAX_COUNT);
            
            meError = meIDS.meQueryNumberSubdevices(idxDevice, out numSubdevices);  
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                int idxSubdevice = 0;
                meIDS.meQueryNameDevice(idxDevice, devName, devName.Capacity);  
                while (idxSubdevice < numSubdevices)
                {
                    meError = meIDS.meQuerySubdeviceByType(idxDevice,                 
                                                           idxSubdevice,              
                                                           subdeviceType,             
                                                           meIDS.ME_SUBTYPE_ANY,      
                                                           out idxMatchedSubdevice);  

                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        AddDiSubdevice(devName.ToString(), idxDevice, idxMatchedSubdevice, subdeviceType);
                        idxSubdevice = idxMatchedSubdevice + 1;  
                    } else {
                        break;     
                    }
                }
            }
        }

        private void AddDiSubdevice(String devName, int idxDevice, int idxMatchedSubdevice, int subdeviceType)
        {
            int numOfChannels;
            meIDS.meQueryNumberChannels(idxDevice, idxMatchedSubdevice, out numOfChannels);   
            subdeviceList.Add(new DiSubdevice(devName, idxDevice, idxMatchedSubdevice, numOfChannels)); 
            
            coBoxSubDevs.Items.Add(devName
                                   + ",  Device " + idxDevice.ToString()
                                   + ",  Subdevice " + idxMatchedSubdevice.ToString()
                                   + ",  Type: " + GetSubDevTypeName(subdeviceType)
                                   + ",  Channels: " + numOfChannels.ToString());
        }

        private void btnSingleRead_Click(object sender, EventArgs e)
        {
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                int meError;
                DiSubdevice diSubdevice = subdeviceList[coBoxSubDevs.SelectedIndex];

                meError = ConfigureSubdeviceForInput(diSubdevice);
                if (meError == meIDS.ME_ERRNO_SUCCESS)
                {
                    int value;
                    meError = ReadSingleValue(diSubdevice, out value);

                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        DisplayValue(diSubdevice, value);
                    }
                    else
                    {
                        DisplayErrorMessage("meIOSingle - Error: ", meError);
                    }
                }
                else
                {
                    DisplayErrorMessage("meIOSingleConfig - Error: ", meError);
                }
            }
        }

        private void DisplayValue(DiSubdevice diSubdevice, int value)
        {
            dataLine++;
            textBoxInfo.AppendText(String.Format("{0:00}", dataLine)
                      + " Read from " + diSubdevice.deviceName
                      + ",  Device: " + diSubdevice.deviceIndex.ToString()
                      + ",  Subdevice: " + diSubdevice.subdevIndex.ToString()
                      + ",  Value hex: " + String.Format("0x{0:X}", value)
                      + ",  bin: " + IntToBin(value, diSubdevice.numOfChannels)
                      + ",  dec: " + value.ToString() + Environment.NewLine);
        }

        private int ReadSingleValue(DiSubdevice diSubdevice, out int value)
        {
            meIDS.meIOSingle_t[] io_single = new meIDS.meIOSingle_t[1];     
            io_single[0].iDevice = diSubdevice.deviceIndex;                 
            io_single[0].iSubdevice = diSubdevice.subdevIndex;              
            io_single[0].iChannel = 0;                                      
            io_single[0].iDir = meIDS.ME_DIR_INPUT;                         
            io_single[0].iValue = 0;                                        
            io_single[0].iTimeOut = 0;                                      
            io_single[0].iFlags = diSubdevice.singleTypeFlag;               
            io_single[0].iErrno = 0;                                        

            int meError = meIDS.meIOSingle(ref io_single[0],
                                       1,                                   //Number of entries in array of meIOSingle_t
                                       meIDS.ME_IO_SINGLE_NO_FLAGS);        
            value = io_single[0].iValue;
            return meError;
        }

        private int ConfigureSubdeviceForInput(DiSubdevice diSubdevice)
        {
            int meError = meIDS.meIOSingleConfig(diSubdevice.deviceIndex,          
                                            diSubdevice.subdevIndex,           
                                            0,                                 
                                            meIDS.ME_SINGLE_CONFIG_DIO_INPUT,  
                                            meIDS.ME_REF_NONE,                 
                                            meIDS.ME_VALUE_NOT_USED,       
                                            meIDS.ME_TRIG_TYPE_NONE,           
                                            meIDS.ME_VALUE_NOT_USED,           
                                            diSubdevice.singleConfigFlag);     
            return meError;
        }


        private void CloseMEiDSDriver()
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
            {
                meIDS.meClose(meIDS.ME_CLOSE_NO_FLAGS);                                                 
                openErrSuccess = Constants.ME_IDS_NOTOPENED;
                btnSingleRead.Enabled = false;
                btnResetSubdev.Enabled = false;
                subdeviceList.Clear();
                coBoxSubDevs.Items.Clear();
            }
        }

        private void btnResetSubdev_Click(object sender, EventArgs e)
        {
            DiSubdevice diSubdev;
            diSubdev = subdeviceList[coBoxSubDevs.SelectedIndex];
            int meError = meIDS.meIOResetSubdevice(diSubdev.deviceIndex,
                                                  diSubdev.subdevIndex,
                                                  meIDS.ME_IO_RESET_SUBDEVICE_NO_FLAGS);
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                textBoxInfo.AppendText("Reset Device " + diSubdev.deviceIndex.ToString() + ", Subdevice " + diSubdev.subdevIndex.ToString() + Environment.NewLine);
            }
            else
            {
                DisplayErrorMessage("Error reset Subdevice", meError);
            }
        }

        private int GetDevicePlugged(int idxDevice, out int plugged)
        {
            int vendorId;
            int deviceId;
            int serialNo;
            int busType;
            int busNo;
            int devNo;
            int funcNo;

            int meError = meIDS.meQueryInfoDevice(idxDevice,
                                             out vendorId,              
                                             out deviceId,              
                                             out serialNo,              
                                             out busType,               
                                             out busNo,                 
                                             out devNo,                 
                                             out funcNo,                
                                             out plugged);              //Indicates whether a device is physically available
            return meError;
        }

        private string GetSubDevTypeName(int subdeviceType)
        {
            string tmpType;
            switch (subdeviceType)
            {
                case meIDS.ME_TYPE_INVALID:
                    tmpType = "ME_TYPE_INVALID";
                    break;
                case meIDS.ME_TYPE_AI:
                    tmpType = "ME_TYPE_AI";
                    break;
                case meIDS.ME_TYPE_AO:
                    tmpType = "ME_TYPE_AO";
                    break;
                case meIDS.ME_TYPE_DIO:
                    tmpType = "ME_TYPE_DIO";
                    break;
                case meIDS.ME_TYPE_DO:
                    tmpType = "ME_TYPE_DO";
                    break;
                case meIDS.ME_TYPE_DI:
                    tmpType = "ME_TYPE_DI";
                    break;
                case meIDS.ME_TYPE_CTR:
                    tmpType = "ME_TYPE_CTR";
                    break;
                case meIDS.ME_TYPE_EXT_IRQ:
                    tmpType = "ME_TYPE_EXT_IRQ";
                    break;
                case meIDS.ME_TYPE_FIO:
                    tmpType = "ME_TYPE_FIO";
                    break;
                case meIDS.ME_TYPE_FO:
                    tmpType = "ME_TYPE_FO";
                    break;
                case meIDS.ME_TYPE_FI:
                    tmpType = "ME_TYPE_FI";
                    break;
                case meIDS.ME_TYPE_FPGA:
                    tmpType = "ME_TYPE_FPGA";
                    break;
                default:
                    tmpType = String.Format("0x{0:X}", subdeviceType);
                    break;
            };
            return tmpType;
        }

        private string IntToBin(int val, int dispDigits)
        {
            int b;
            bool found = false;
            string tmpStr = "";
            for (b = 0; b < Constants.DIGITNUMS.Length; b++)
            {
                if (dispDigits == Constants.DIGITNUMS[b])
                {
                    found = true;
                    break;
                }
            }
            if (found)
            {
                if (dispDigits == 1)
                {
                    tmpStr = (val & 1).ToString();
                }
                else
                {
                    for (int i = (dispDigits >> 2) - 1; i >= 0; i--)
                    {
                        b = (val >> (i << 2)) & 0x0F;
                        tmpStr += Constants.BIN_STR[b];
                    }
                }
            }
            return tmpStr;
        }

        private void DisplayVersionMainDriver()
        {
            int tmpVersion;
            meIDS.meQueryVersionMainDriver(out tmpVersion);
            textBoxInfo.AppendText("Maindriver version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
        }

        private void DisplayVersionLibrary()
        {
            int tmpVersion;
            meIDS.meQueryVersionLibrary(out tmpVersion);
            textBoxInfo.AppendText("Library version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
        }

        private void DisplayErrorMessage(string message, int meError)
        {
            StringBuilder errMessage = new StringBuilder(meIDS.ME_ERROR_MSG_MAX_COUNT);
            textBoxInfo.AppendText(message + meError.ToString() + Environment.NewLine);
            meIDS.meErrorGetMessage(meError, errMessage, errMessage.Capacity);             
            textBoxInfo.AppendText("ErrorMessage: " + errMessage + Environment.NewLine);   
        }

        private void btnClearHwInfo_Click(object sender, EventArgs e)
        {
            textBoxInfo.Clear();
            dataLine = 0;
        }

        private string MeVersionToString(int version)
        {
            return String.Format("{0}.{1}.{2}.{3}", version >> 24, 0xFF & (version >> 16), 0xFF & (version >> 8), 0xFF & version);
        }

        private void btnAbout_Click(object sender, EventArgs e)
        {
            AboutBox1 aboutBox = new AboutBox1();
            aboutBox.ShowDialog();
        }
    }

    class DiSubdevice
    {
        public string deviceName;
        public int deviceIndex;
        public int subdevIndex;
        public int numOfChannels;
        public int singleConfigFlag;
        public int singleTypeFlag;
        public DiSubdevice(string DeviceName, int DeviceIndex, int SubdevIndex, int NumOfChannels)
        {
            this.deviceName = DeviceName;
            this.deviceIndex = DeviceIndex;
            this.subdevIndex = SubdevIndex;
            this.numOfChannels = NumOfChannels;
            SetSingleFlags(NumOfChannels);
        }

        private void SetSingleFlags(int NumberOfChannels)
        {
            switch (NumberOfChannels)
            {
                case 1:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_BIT;    
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_BIT;
                    break;
                case 8:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_BYTE;   
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_BYTE;
                    break;
                case 16:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_WORD;   
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_WORD;
                    break;
                case 24:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_DWORD;  
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_DWORD;
                    break;
                case 32:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_DWORD;  
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_DWORD;
                    break;
                default:
                    singleConfigFlag = meIDS.ME_VALUE_INVALID;  
                    singleTypeFlag = meIDS.ME_VALUE_INVALID;
                    break;
            };
        }
    }

    static class Constants
    {
        public const int ME_IDS_NOTOPENED = -1;
        public static readonly string[] BIN_STR = {"0000", "0001", "0010", "0011",
                                                   "0100", "0101", "0110", "0111",
                                                   "1000", "1001", "1010", "1011",
                                                   "1100", "1101", "1110", "1111"};
        public static int[] DIGITNUMS = { 1, 8, 16, 24, 32 }; 
    }
}
